import torch
import numpy as np
from tqdm import tqdm

def _edge_mask(labels, ignore=[None]):
    labels = labels.squeeze()
    first_row = labels[0, :]
    last_row = labels[-1, :]
    first_column = labels[:, 0]
    last_column = labels[:, -1]

    edges = []
    if 'top' not in ignore:
        edges.append(first_row)
    if 'bottom' not in ignore:
        edges.append(last_row)
    if 'left' not in ignore:
        edges.append(first_column)
    if 'right' not in ignore:
        edges.append(last_column)

    if len(edges) == 0:
        return torch.zeros_like(labels).bool()

    edges = torch.cat(edges, dim=0)
    return torch.isin(labels, edges[edges > 0])


def _remove_edge_labels(labels, ignore=[None]):
    return labels * ~_edge_mask(labels, ignore=ignore)



def _chops(img_shape: tuple, shape: tuple, overlap: int = 0) -> tuple:
    """This function splits an image into desired windows and returns the indices of the windows"""

    if (torch.tensor(img_shape[-2:]) < torch.tensor(shape)).any():
        return [0], [0]
    h, v = img_shape[-2:]

    assert shape[0] > overlap and shape[1] > overlap, f"The overlap {overlap} must be smaller than the window size {shape}"

    stride_h = shape[0] - overlap
    stride_v = shape[1] - overlap

        
    max_v = (v - shape[1])
    max_h = (h - shape[0])

    v_index = np.array([i * stride_v for i in range(v // stride_v )] + [v - shape[1]])
    h_index = np.array([i * stride_h for i in range(h // stride_h )] + [h - shape[0]])

    v_index = v_index[v_index <= max_v]
    h_index = h_index[h_index <= max_h]

    v_index = np.unique(v_index)
    h_index = np.unique(h_index)

    return h_index, v_index


def _tiles_from_chops(image: torch.Tensor, shape: tuple, tuple_index: tuple) -> list:
    """This function takes an image, a shape, and a tuple of window indices (e.g., output of the function _chops)
    and returns a list of windows"""
    h_index, v_index = tuple_index

    if len(image.shape) == 2:
        image = image.unsqueeze(0)

    stride_h = shape[0]
    stride_v = shape[1]
    tile_list = []
    for i, window_i in enumerate(h_index):
        for j, window_j in enumerate(v_index):
            current_window = image[..., window_i:window_i + stride_h, window_j:window_j + stride_v]
            tile_list.append(current_window)
    return tile_list


def _stitch(tiles: list, shape: tuple, chop_list: list, final_shape: tuple, offset : int):
    """This function takes a list of tiles, a shape, and a tuple of window indices (e.g., outputed by the function _chops)
    and returns a stitched image"""
    from instanseg.utils.pytorch_utils import torch_fastremap, match_labels

    canvas = torch.zeros(final_shape, dtype=torch.int32, device = tiles[0].device)

    running_max = 0
            
    for i, window_i in enumerate(chop_list[0]):
        for j, window_j in enumerate(chop_list[1]):

            edge_window = False
            new_tile = tiles[i * len(chop_list[1]) + j][None]
            ignore_list = []
            if i == 0:
                ignore_list.append("top")
            if j == 0:
                ignore_list.append("left")
            if i == len(chop_list[0])-1:
                ignore_list.append("bottom")
            if j == len(chop_list[1])-1:
                ignore_list.append("right")

            if i == len(chop_list[0])-1 and j == len(chop_list[1])-1:
                edge_window = True
                tile1 = canvas[..., window_i + offset:window_i + shape[0], window_j + offset:window_j + shape[1]]
                tile2 = _remove_edge_labels(new_tile[:,offset:shape[0],offset: shape[1]], ignore = ignore_list)

            elif i == len(chop_list[0])-1:
                edge_window = True
                tile1 = canvas[..., window_i + offset:window_i + shape[0], window_j + offset:window_j + shape[1]]
                tile2 = _remove_edge_labels(new_tile[:,offset:shape[0],offset : shape[1]], ignore =ignore_list)

            elif j == len(chop_list[1])-1:
                edge_window = True
                tile1 = canvas[..., window_i + offset:window_i + shape[0], window_j + offset:window_j + shape[1]]
                tile2 = _remove_edge_labels(new_tile[:,offset:shape[0],offset: shape[1]], ignore = ignore_list)

            if i == 0 and j == 0:
                edge_window = True
                tile1 = canvas[..., window_i  :window_i + shape[0], window_j :window_j + shape[1]]
                tile2 = _remove_edge_labels(new_tile[:, :shape[0], : shape[1]], ignore = ignore_list)
            elif i == 0:
                edge_window = True
                tile1 = canvas[..., window_i  :window_i + shape[0], window_j + offset :window_j + shape[1]]
                tile2 = _remove_edge_labels(new_tile[:, :shape[0],offset : shape[1]], ignore = ignore_list)

            elif j == 0:
                edge_window = True
                tile1 = canvas[..., window_i  + offset:window_i + shape[0], window_j:window_j + shape[1]]
                tile2 = _remove_edge_labels(new_tile[:,offset :shape[0],: shape[1]], ignore = ignore_list)

            if edge_window:

                tile2 = torch_fastremap(tile2)
                tile2[tile2>0] = tile2[tile2>0] + running_max

                remapped = match_labels(tile1, tile2, threshold = 0.1)[1]
                tile1[remapped>0] = remapped[remapped>0].int()

                running_max = max(running_max, tile1.max())

            else:
                
                tile1 = canvas[..., window_i + offset:window_i + shape[0] - offset, window_j + offset:window_j + shape[1] - offset]
                tile2 = _remove_edge_labels(new_tile[:,offset:shape[0] -offset,offset: shape[1]-offset])


                tile2 = torch_fastremap(tile2)
                tile2[tile2>0] = tile2[tile2>0] + running_max

                remapped = match_labels(tile1, tile2, threshold = 0.1)[1]
                tile1[remapped>0] = remapped[remapped>0].int()

                running_max = max(running_max, tile1.max())
            

    return canvas


def _zarr_to_json_export(path_to_zarr, detection_size = 30, size = 1024, scale = 1, n_dim = 1):

    import zarr
    import numpy as np
    import os

    z = zarr.open(path_to_zarr, mode='r')

    output_path = str(path_to_zarr).replace(".zarr",".geojson")   


    from instanseg.utils.tiling import _chops

    chop_list = _chops(z.shape, shape=(size,size), overlap=detection_size)

    from instanseg.utils.utils import labels_to_features
    from instanseg.utils.tiling import _remove_edge_labels
    import json

    count = 0
    #Delete the file if it already exists
    if os.path.exists( output_path):
        os.remove( output_path)

    with open(os.path.join(output_path), "a") as outfile:
        outfile.write('[')
        outfile.write('\n')

    if n_dim == 1:
        classes = ["Detection"]
    else:
        classes = ["Nucleus","Cell"]


    for i, window_i in tqdm(enumerate(chop_list[0]), total = len(chop_list[0])):
        for j, window_j in enumerate(chop_list[1]):
            for n in range(n_dim):
                
                image = z[n,window_i:window_i+size, window_j:window_j+size]

                # if image.max() > 0:
                #     from instanseg.utils.utils import show_images
                #     show_images(image)
              
                image = _remove_edge_labels(torch.tensor(image)).numpy()

                features = labels_to_features(image.astype(np.int32), object_type='detection', include_labels=True,
                                        classification=classes[n],offset=[window_j*scale,window_i*scale], downsample = scale)
                
                features = features["features"]

        #    print(features)
            

            if len(features) > 0:
                count+=1
                geojson = json.dumps(features)

                with open(os.path.join( output_path), "a") as outfile:
                    outfile.write(geojson[1:-1] + ",")
                    outfile.write('\n')

    with open(os.path.join( output_path), "a") as outfile:
        outfile.write(']')


def _instanseg_padding(img: torch.Tensor, extra_pad: int = 0, min_dim: int = 16, ensure_square: bool = False):

    is_square = img.shape[-2] == img.shape[-1]
    original_shape = img.shape[-2:]
    bigger_dim = max(img.shape[-2], img.shape[-1])

    if ensure_square and not is_square:
        img = torch.functional.F.pad(img, [0, bigger_dim - img.shape[-1], 0, bigger_dim - img.shape[-2]], mode='constant')

    padx = min_dim * torch.ceil(torch.tensor((img.shape[-2] / min_dim))).int() - img.shape[-2] + extra_pad * 2
    pady = min_dim * torch.ceil(torch.tensor((img.shape[-1] / min_dim))).int() - img.shape[-1] + extra_pad * 2

    if padx > img.shape[-2]:
        padx = padx - extra_pad
    if pady > img.shape[-1]:
        pady = pady - extra_pad
    img = torch.functional.F.pad(img, [0, int(pady), 0, int(padx)], mode='reflect')

    if ensure_square and not is_square:
        pady = pady + bigger_dim - original_shape[-1]
        padx = padx + bigger_dim - img.shape[-2]
        print(padx, pady)
    

    return img, torch.stack((padx, pady))


def _recover_padding(x: torch.Tensor, pad: torch.Tensor):
    # x must be 1,C,H,W or C,H,W
    squeeze = False
    if x.ndim == 3:
        x = x.unsqueeze(0)
        squeeze = True

    if pad[0] == 0:
        pad[0] = -x.shape[2]
    if pad[1] == 0:
        pad[1] = -x.shape[3]

    if squeeze:
        return x[:, :, :-pad[0], :-pad[1]].squeeze(0)
    else:
        return x[:, :, :-pad[0], :-pad[1]]



def _sliding_window_inference(input_tensor, 
                             predictor, 
                             window_size=(512, 512), 
                             overlap = 80, 
                             max_cell_size = 20, 
                             sw_device='cuda',
                             device='cpu', 
                             output_channels=1, 
                             show_progress = True, 
                             batch_size = 1,
                             **kwargs):
    
    h,w = input_tensor.shape[-2:]
    window_size = (min(window_size[0], h), min(window_size[1], w))
    
    input_tensor = input_tensor.to(device)
    predictor = predictor.to(sw_device)
 
    tuple_index = _chops(input_tensor.shape, shape=window_size, overlap=2 * (overlap + max_cell_size))
    tile_list = _tiles_from_chops(input_tensor, shape=window_size, tuple_index=tuple_index)
 
 
    assert len(tile_list) > 0, "No tiles generated"
    # print("Number of tiles: ", len(tile_list))
    # print("Shape of tiles: ", tile_list[0].shape)
    # print("window size: ", window_size)
    # print("input tensor shape: ", input_tensor.shape)
 
    with torch.no_grad():
        with torch.amp.autocast("cuda"):
            batch_list = [torch.stack(tile_list[batch_size * i:batch_size * (i+1)]) for i in range(int(np.ceil(len(tile_list)/batch_size)))]
            label_list = torch.cat([predictor(tile.to(sw_device),**kwargs).to(device) for tile in tqdm(batch_list, disable= not show_progress,leave = False, colour = "blue")])
 
   
    lab = torch.cat([_stitch([lab[i] for lab in label_list],
                            shape=window_size,
                            chop_list=tuple_index,
                            offset = overlap,
                            final_shape=(1, input_tensor.shape[1], input_tensor.shape[2])) 
 
                    for i in range(output_channels)], dim=0)
 
 
    return lab[None]  # 1,C,H,W




import os
import bioimageio.core
import numpy as np
import torch
from bioio import BioImage

from instanseg.utils.augmentations import Augmentations
from instanseg.utils.utils import _choose_device, show_images
from instanseg.utils.model_loader import load_model


def set_export_paths():
    from pathlib import Path
    if os.environ.get('INSTANSEG_BIOIMAGEIO_PATH'):
        path = Path(os.environ['INSTANSEG_BIOIMAGEIO_PATH'])
    else:
        path = Path(os.path.join(os.path.dirname(__file__),"../bioimageio_models/"))
        os.environ['INSTANSEG_BIOIMAGEIO_PATH'] = str(path)

    if not path.exists():
        path.mkdir(exist_ok=True,parents=True)

    if os.environ.get('INSTANSEG_TORCHSCRIPT_PATH'):
        path = Path(os.environ['INSTANSEG_TORCHSCRIPT_PATH'])
    else:
        path = Path(os.path.join(os.path.dirname(__file__),"../torchscripts/"))
        os.environ['INSTANSEG_TORCHSCRIPT_PATH'] = str(path)

    if not path.exists():
        path.mkdir(exist_ok=True,parents=True)

    if os.environ.get('INSTANSEG_MODEL_PATH'):
        path = Path(os.environ['INSTANSEG_MODEL_PATH'])
    else:
        path = Path(os.path.join(os.path.dirname(__file__),"../models/"))
        os.environ['INSTANSEG_MODEL_PATH'] = str(path)

    if not path.exists():
        path.mkdir(exist_ok=True,parents=True)


dataset_dict = {
    "DSB_2018": ["CC 0","https://bbbc.broadinstitute.org/BBBC038"],
    "CoNSeP": ["Apache 2.0","https://warwick.ac.uk/fac/cross_fac/tia/data/hovernet"],
    "TNBC_2018": ["CC BY 4.0","https://zenodo.org/records/3552674"],
    "MoNuSeg": ["CC BY NC 4.0","https://monuseg.grand-challenge.org/"],
    "LyNSec": ["CC BY 4.0","https://zenodo.org/records/8065174"],
    "LyNSeC": ["CC BY 4.0","https://zenodo.org/records/8065174"],
    "NuInsSeg": ["CC BY 4.0","https://zenodo.org/records/10518968"],
    "IHC_TMA": ["CC BY 4.0","https://zenodo.org/records/7647846"],
    "CPDMI_2023": ["CC BY 4.0","https://www.nature.com/articles/s41597-023-02108-z"],
    "cellpose": ["NC","https://www.cellpose.org/dataset"],
    "TissueNet": ["Modified Apache, Non-Commercial", "https://datasets.deepcell.org/"],
    "CIL": ["CC BY 3.0", "https://www.cellimagelibrary.org/images/CCDB_6843"],
    "BSST265": ["CC0","https://www.ebi.ac.uk/biostudies/bioimages/studies/S-BSST265"],
}


from bioimageio.spec.model.v0_5 import ModelDescr

def readme(model_name: str, model_dict: dict = None):
    # create markdown documentation for your model
    # this should describe how the model was trained, (and on which data)
    # and also what to take into consideration when running the model, especially how to validate the model
    # here, we just create a stub documentation

    print("Creating README file at", os.path.join(model_name, model_name + "_README.md"))

    with open(os.path.join(model_name, model_name + "_README.md"), "w") as f:
        f.write("# This is an InstanSeg model. \n")
        f.write("The InstanSeg method is shared with an Apache-2.0 license.\n\n")
        f.write("""For an introduction & comparison to other approaches for nucleus segmentation in brightfield histology images, see: \n > Goldsborough, T. et al. (2024) InstanSeg: an embedding-based instance segmentation algorithm optimized for accurate, efficient and portable cell segmentation. _arXiv_. Available at: https://doi.org/10.48550/arXiv.2408.15954. \n\n To read about InstanSeg's extension to nucleus + full cell segmentation and support for fluorescence & multiplexed images, see: \n > Goldsborough, T. et al. (2024) A novel channel invariant architecture for the segmentation of cells and nuclei in multiplexed images using InstanSeg. _bioRxiv_, p. 2024.09.04.611150. Available at: https://doi.org/10.1101/2024.09.04.611150. \n""")
        f.write("\n This model was trained on the following datasets: \n")
        
        if model_dict is not None and "source_dataset" in model_dict.keys():
            for dataset in (model_dict["source_dataset"]).replace("[","").replace("]","").replace("'","").split(", "):
                if dataset in dataset_dict.keys():
                    f.write(f"- {dataset} \n")
                    f.write(f"  - License: {dataset_dict[dataset][0]} \n")
                    f.write(f"  - URL: {dataset_dict[dataset][1]} \n")
                else:
                    f.write(f"- {dataset} \n")
                    f.write(f"  - License: Not specified \n")
                    f.write(f"  - URL: Not specified \n")
        

        f.write("\n The user is responsible for ensuring that the model is used in accordance with the licenses of the source datasets. \n")

          #  f.write(str(model_dict["source_dataset"]))

import os, shutil
def make_archive(source, destination):
        base = os.path.basename(destination)
        name = base.split('.')[0]
        format = base.split('.')[1]
        archive_from = os.path.dirname(source)
        archive_to = os.path.basename(source.strip(os.sep))
        shutil.make_archive(name, format, root_dir=archive_from, base_dir=archive_to)
        shutil.move('%s.%s'%(name,format), destination)


def export_bioimageio(torchsript: torch.jit._script.RecursiveScriptModule, 
                      model_name: str, 
                      test_img_path: str, 
                      model_dict: dict = None, 
                      output_name = None,
                      output_channel_names = None,
                      output_types = ["instance_segmentation"],
                      version: str = None):
    
    set_export_paths()

    output_path = os.environ['INSTANSEG_BIOIMAGEIO_PATH']

    if output_name is None:
        output_name = model_name
    # create a directory to store bioimage.io model files
    os.makedirs(output_name, exist_ok=True)
    # save the model weights
    torchsript.save(os.path.join(output_name, "instanseg.pt"))


    try:
        model,model_dict = load_model(model_name, path = os.environ['INSTANSEG_MODEL_PATH'])
    except:
        raise Exception("Model configuration files could not be loaded")
    
    print("Model pixel size: ", model_dict["pixel_size"])
    model_pixel_size = model_dict["pixel_size"]

    torchsript.eval()
    device = _choose_device()
    torchsript.to(device)

    img = BioImage(test_img_path)
    if "S" in img.dims.order and img.dims.S > img.dims.C:
        input_data = img.get_image_data("SYX")
    else:
        input_data = img.get_image_data("CYX")
        
    if img.physical_pixel_sizes.X is not None:
        pixel_size = img.physical_pixel_sizes.X
        print("Pixel size was found in the metadata, pixel size is set to: ", pixel_size)
    else:
        pixel_size = 0.5
        print("Pixel size was not found in the metadata, please set the pixel size of the input image in microns manually")

    if model_dict["channel_invariant"]:
        dim_in = 1
        step = 1
    else:
        dim_in = model_dict["dim_in"]
        step = 0


    target_segmentation = model_dict["target_segmentation"]
    print("Cells and nuclei: ", target_segmentation)
    if output_channel_names is None:
        if target_segmentation == "N":
            output_channel_names = ["nuclei"]
        elif target_segmentation == "C":
            output_channel_names = ["cells"]
        elif target_segmentation == "NC":
            output_channel_names = ["nucei", "cells"]
        else:
            output_channel_names = ["nuclei"]
        print("Assuming output channel names: ", output_channel_names)

    Augmenter=Augmentations()

    input_tensor,_ = Augmenter.to_tensor(input_data,normalize=False) #this converts the input data to a tensor and does percentile normalization (no clipping)
    import math
    if math.isnan(model_pixel_size):
        model_pixel_size_tmp = pixel_size
    else:
        model_pixel_size_tmp = model_pixel_size
    input_crop,_ = Augmenter.torch_rescale(input_tensor,labels=None,current_pixel_size=pixel_size,requested_pixel_size=model_pixel_size_tmp,crop = True, random_seed=1)
    input_crop = input_crop.unsqueeze(0) # add batch dimension
    if input_crop.shape[1] != dim_in and not model_dict["channel_invariant"]:
        input_crop = torch.zeros((1,dim_in,input_crop.shape[2],input_crop.shape[3]),dtype=torch.float32, device = input_crop.device)

    print("Input tensor shape: ", input_crop.shape)

    np.save(os.path.join(output_name, "test-input.npy"), input_crop.float().numpy())

    input_crop,_ = Augmenter.to_tensor(input_crop[0],normalize=True)
    input_crop = input_crop.unsqueeze(0)

    with torch.no_grad():
        output = torchsript(input_crop.to(device))

        if isinstance(output, tuple):
            assert len(output) == len(output_types)
            for i, out in enumerate(output):
                np.save(os.path.join(output_name, f"test-output_{output_types[i]}.npy"), out.cpu().numpy())
        else:
            np.save(os.path.join(output_name, f"test-output_{output_types[0]}.npy"), output.cpu().numpy())

    from instanseg.utils.utils import _display_overlay

    if isinstance(output, tuple):
        output = output[0]

    cover = _display_overlay(input_crop[0], output)
    show_images(cover, colorbar=False)
    show_images(cover, colorbar=False, save_str= os.path.join(output_name, "cover"))

    if model_dict is not None and "source_dataset" in model_dict.keys():
        train_data = str(model_dict["source_dataset"])
    else:
        train_data = "Not specified"

    # create readme
    readme(output_name, model_dict)

    if not os.path.exists(output_path):
        os.makedirs(output_path)


    if os.path.exists(os.path.join(output_path, output_name + ".zip")):
        print("Removing existing model zip")
        os.remove(os.path.join(output_path, output_name + ".zip"))

    from bioimageio.spec.model.v0_5 import (
        AxisId,
        BatchAxis,
        ChannelAxis,
        FileDescr,
        Identifier,
        InputTensorDescr,
        IntervalOrRatioDataDescr,
        ParameterizedSize,
        SpaceInputAxis,
        SpaceOutputAxis,
        IndexInputAxis,
        TensorId,
        WeightsDescr,
        ScaleRangeDescr,
    )

    if model_dict["channel_invariant"]:
        input_axes = [BatchAxis(), IndexInputAxis( description = "Channel axis for channel invariant models", size=ParameterizedSize(min=1, step=1))]
    else:
        input_axes = [BatchAxis(), ChannelAxis(id = AxisId("channel"),channel_names=[Identifier(i) for i in ["C1","C2","C3"]])]

    if input_crop.ndim == 5: 
        input_axes += [
            SpaceInputAxis(id=AxisId("z"), size=ParameterizedSize(min=32, step=1)),
            SpaceInputAxis(id=AxisId("y"), size=ParameterizedSize(min=32, step=1)),
            SpaceInputAxis(id=AxisId("x"), size=ParameterizedSize(min=32, step=1)),
        ]
        data_descr = IntervalOrRatioDataDescr(type="float32")
    elif input_crop.ndim == 4: 
        input_axes += [
            SpaceInputAxis(id=AxisId("y"), size=ParameterizedSize(min=32, step=1),scale = model_pixel_size_tmp,unit = "micrometer",),

            SpaceInputAxis(id=AxisId("x"), size=ParameterizedSize(min=32, step=1),scale = model_pixel_size_tmp,unit = "micrometer",),
        ]
        data_descr = IntervalOrRatioDataDescr(type="float32")
    else:
        raise NotImplementedError(
            f"Recreating inputs for {input_crop.shape} is not implemented"
        )
    
    preprocessing = [ScaleRangeDescr(kwargs = {"min_percentile": 0.1, "max_percentile": 99.9, "eps": 1e-6, "axes": ['x','y']})]

    input_descr = InputTensorDescr(
        id=TensorId("raw"),
        axes=input_axes,
        test_tensor=FileDescr(source=os.path.join(output_name, "test-input.npy"),),
        data=data_descr,
        preprocessing=preprocessing,

    )

    def get_instance_segmentation_descriptor():
        from bioimageio.spec.model.v0_5 import OutputTensorDescr, SizeReference

        output_axes = [
            BatchAxis(),
            ChannelAxis(
                channel_names=[Identifier(n) for n in output_channel_names]
            ),
        ]

        if output.ndim == 5: 
            output_axes += [
                SpaceOutputAxis(
                    id=AxisId("z"),
                    size=SizeReference(tensor_id=TensorId("raw"), axis_id=AxisId("z")),
                    scale=model_pixel_size_tmp,
                    unit="micrometer",
                ),  # same size as input (tensor `raw`) axis `z`
                SpaceOutputAxis(
                    id=AxisId("y"),
                    size=SizeReference(tensor_id=TensorId("raw"), axis_id=AxisId("y")),
                    scale=model_pixel_size_tmp,
                    unit="micrometer",
                ),
                SpaceOutputAxis(
                    id=AxisId("x"),
                    size=SizeReference(tensor_id=TensorId("raw"), axis_id=AxisId("x")),
                    scale=model_pixel_size_tmp,
                    unit="micrometer",
                ),
            ]
        elif output.ndim == 4:
            output_axes += [
                SpaceOutputAxis(
                    id=AxisId("y"),
                    size=SizeReference(tensor_id=TensorId("raw"), axis_id=AxisId("y")),
                    scale=model_pixel_size_tmp,
                    unit="micrometer",
                ),  # same size as input (tensor `raw`) axis `y`
                SpaceOutputAxis(
                    id=AxisId("x"),
                    size=SizeReference(tensor_id=TensorId("raw"), axis_id=AxisId("x")),
                    scale=model_pixel_size_tmp,
                    unit="micrometer",
                ),
            ]
        else:
            raise NotImplementedError(
                f"Recreating outputs for { output.ndim} is not implemented"
            )

        output_descr = OutputTensorDescr(
            id=TensorId("instance_segmentation"),
            axes=output_axes,
            test_tensor=FileDescr(source=os.path.join(output_name, "test-output_instance_segmentation.npy"),),
        )

        return output_descr
    

    def get_detection_classes_descriptor():

        from bioimageio.spec.model.v0_5 import IndexOutputAxis, DataDependentSize, OutputTensorDescr,NominalOrOrdinalDataDescr

        output_axes = [
           # BatchAxis(),
            IndexOutputAxis(id = "n_objects", 
                            description = "Number of detected objects",
                            size = DataDependentSize()),
            IndexOutputAxis(id = "n_classes",
                            description = "Number of classes",
                            size = 1,),
            
        ]

        output_descr = OutputTensorDescr(
            id=TensorId("detection_classes"),
            axes=output_axes,
            test_tensor=FileDescr(source=os.path.join(output_name, "test-output_detection_classes.npy"),),
            data = NominalOrOrdinalDataDescr(type = "uint8",values = ["fancy_class_" + str(i) for i in range(19)]),
        )

        return output_descr
    

    def get_detection_logits_descriptor():

        from bioimageio.spec.model.v0_5 import IndexOutputAxis, DataDependentSize, OutputTensorDescr,NominalOrOrdinalDataDescr

        output_axes = [
           # BatchAxis(),
            IndexOutputAxis(id = "n_objects", 
                            description = "Number of detected objects",
                            size = DataDependentSize()),
            IndexOutputAxis(id = "logits",
                            description = "Logits for classes",
                            size = 19,),
            
        ]

        output_descr = OutputTensorDescr(
            id=TensorId("detection_logits"),
            axes=output_axes,
            test_tensor=FileDescr(source=os.path.join(output_name, "test-output_detection_classes.npy"),),
            data = NominalOrOrdinalDataDescr(type = "float32", values = [0]),
        )

        return output_descr
    
    outputs = []
    for output_type in output_types:
        if output_type == "instance_segmentation":
            outputs.append(get_instance_segmentation_descriptor())
        elif output_type == "detection_classes":
            outputs.append(get_detection_classes_descriptor())
        elif output_type == "detection_logits":
            outputs.append(get_detection_logits_descriptor())
        else:
            raise NotImplementedError(f"Output type {output_type} is not implemented")


    from bioimageio.spec.model.v0_5 import (
        Author,
        CiteEntry,
        Doi,
        HttpUrl,
        LicenseId,
        TorchscriptWeightsDescr,
        RelativeFilePath,
    )
    print(__file__)

    my_model_descr = ModelDescr(
            name=output_name,
            version =version,
            description=f"An InstanSeg model trained for instance segmentation. This model originates from model id: {model_name}",
            authors=[
                Author(name="Thibaut Goldsborough", affiliation="School of Informatics, University of Edinburgh", github_user="ThibautGoldsborough")
            ],
            covers = [os.path.join(output_name, "cover.png")],
            attachments=[FileDescr(source = (os.path.join(os.path.dirname(__file__),"rdf_scripts/instanseg_preprocess.ijm"))),
                         FileDescr(source = (os.path.join(os.path.dirname(__file__),"rdf_scripts/instanseg_postprocess.ijm"))),],

            cite=[
                CiteEntry(text="If you use InstanSeg for nucleus segmentation of brightfield histology images, please cite:", doi=Doi("10.48550/arXiv.2408.15954")),
                CiteEntry(text="If you use InstanSeg for nucleus and/or cell segmentation in fluorescence images, please cite:", doi=Doi("10.1101/2024.09.04.611150"))
            ],
            license=LicenseId("Apache-2.0"),
            documentation=RelativeFilePath(os.path.join(output_name, output_name + "_README.md"))
            ,
            git_repo=HttpUrl(
                "https://github.com/instanseg/instanseg"
            ), 
            inputs=[input_descr], 
            outputs=outputs, 
            weights=WeightsDescr(
            torchscript=TorchscriptWeightsDescr(
                source=os.path.join(output_name, "instanseg.pt"),
                pytorch_version="2.0.0",
            ),
        ),
        )
    print(f"created '{my_model_descr.name}'")

    from pathlib import Path

    from bioimageio.spec import save_bioimageio_package

    print(
        "package path:",
        save_bioimageio_package(my_model_descr, output_path=Path(os.path.join(output_path, output_name + ".zip"))),
    )

    from bioimageio.core import test_model

    summary = test_model(my_model_descr)
    summary.display()

    # #Cleanup 

    # files_to_remove = [
    #     "cover.png",
    #     "test-output.npy",
    #     "test-input.npy",
    #     "instanseg.pt",
    #     output_name + "_README.md",
    #     "sample_output_0.tif",
    #     "sample_input_0.tif"
    # ]

    # for file in files_to_remove:
    #     if os.path.exists(file):
    #         os.remove(file)

    # #unzip the folder

    # import zipfile
    # import shutil

    # input = os.path.join(output_path, output_name + ".zip")
    # destination = os.path.join(output_path, output_name)
    # with zipfile.ZipFile(input, 'r') as zip_ref:
    #     zip_ref.extractall(destination)
    
    # yaml_path = os.path.join(destination, 'rdf.yaml')


    # #copy ijm files
    import shutil
    # shutil.copyfile(os.path.join(os.path.dirname(__file__),"./rdf_scripts/instanseg_preprocess.ijm"), os.path.join(os.path.dirname(yaml_path), "instanseg_preprocess.ijm"))
    # shutil.copyfile(os.path.join(os.path.dirname(__file__),"./rdf_scripts/instanseg_postprocess.ijm"), os.path.join(os.path.dirname(yaml_path), "instanseg_postprocess.ijm"))


    # make_archive(destination, input)

    shutil.rmtree(output_name)



import torch
import torch
import numpy as np
import matplotlib.pyplot as plt
from tqdm.auto import tqdm
from instanseg.utils.metrics import _robust_average_precision, _robust_f1_mean_calculator

from instanseg.utils.augmentations import Augmentations
import time

from instanseg.utils.utils import show_images
import warnings


global_step = 0
def train_epoch(train_model, 
                train_device, 
                train_dataloader, 
                train_loss_fn, 
                train_optimizer, 
                args,
                ):
    
    global global_step
    start = time.time()
    train_model.train()
    train_loss = []
    for image_batch, labels_batch, _ in tqdm(train_dataloader, disable=args.on_cluster):

        image_batch = image_batch.to(train_device)
        labels = labels_batch.to(train_device)
        output = train_model(image_batch)
        loss = train_loss_fn(output, labels.clone()).mean()
        train_optimizer.zero_grad()
        loss.backward()

        torch.nn.utils.clip_grad_norm_(train_model.parameters(), args.clip)

        train_optimizer.step()
        train_loss.append(loss.detach().cpu().numpy())

    end = time.time()

    return np.mean(train_loss), end - start
    

global_step_test = 0
def test_epoch(test_model, 
               test_device, 
               test_dataloader, 
               test_loss_fn, 
               args,
               postprocessing_fn,
               method,
               iou_threshold,
               debug=False, 
               save_str=None, 
               save_bool=False,
               best_f1=None):
    global global_step_test
    start = time.time()

    test_model.eval()
    test_loss = []

    current_f1_list = []
    with torch.no_grad():
        for image_batch, labels_batch, _ in tqdm(test_dataloader, disable=args.on_cluster):
            image_batch = image_batch.to(test_device)
            labels = labels_batch.to(test_device) 
            output = test_model(image_batch)  
            loss = test_loss_fn(output, labels.clone()).mean()
            test_loss.append(loss.detach().cpu().numpy())

            if type(output) == list:
                output = output[0]


            if labels.type() != 'torch.cuda.FloatTensor' and labels.type() != 'torch.FloatTensor':
                predicted_labels = torch.stack([postprocessing_fn(out) for out in output])
                f1i = _robust_average_precision(labels.clone(), predicted_labels.clone(),
                                               threshold=iou_threshold)

                current_f1_list.append((f1i))
            else:
                warnings.warn("Labels are of type float, not int. Not calculating F1.")
                current_f1_list.append(0)

            global_step_test += 1

    f1_array = np.array(current_f1_list)  # either N,2 or N,

    if f1_array.ndim == 1:
        f1_array = np.atleast_2d(f1_array).T

    mean1_f1 = np.nanmean(f1_array, axis=0)

    mean_f1 = _robust_f1_mean_calculator(mean1_f1)
    #  mean_f1 = current_f1_list

    if mean_f1 > best_f1 or save_bool:
        if len(image_batch[0]) == 3:
            input1 = image_batch[0]
        else:
            input1 = image_batch[0][0]
        labels_dst = labels[0]
        lab = postprocessing_fn(output[0])

        if lab.squeeze().dim() == 2:
            show_images([input1] + [label_i for label_i in labels_dst] + [lab] + [out for out in output[0]],
                        save_str=save_str,
                        titles=["Source"] + ["Label" for _ in labels_dst] + ["Prediction"] + ["Out" for _ in output[0]],
                        labels=[1, 2])
        else:
            show_images([input1] + [label_i for label_i in labels_dst] + [label_i for label_i in lab] + [out for out in
                                                                                                         output[0]],
                        save_str=save_str,
                        titles=["Source"] + ["Label: Nuclei", "Label: Cells"] + ["Prediction: Nuclei",
                                                                                 "Prediction: Cells"] + ["Out" for _ in
                                                                                                         output[0]],
                        labels=[1, 2, 3, 4], n_cols=5)

    end = time.time()
    return np.mean(test_loss), mean1_f1, end - start


def collate_fn(data):
    # data is of length batch size
    # data[0][0] is first image, data[0][1] os the first label

    # print(data[0][0].shape,len(data))
    """
       data: is a list of tuples with (example, label, length)
             where 'example' is a tensor of arbitrary shape
             and label/length are scalars
    """
    imgs, labels = zip(*data)
    lengths = [img.shape[0] for img in imgs]

    max_len = max(lengths)
    C, H, W = data[0][0].shape
    images = torch.zeros((len(data), max_len, H, W))
    labels = torch.stack(labels)
    lengths = torch.tensor(lengths)

    for i, img in enumerate(imgs):
        images[i, :len(img)] = img

    return images, labels, lengths.int()


# import fastremap
class Segmentation_Dataset():
    def __init__(self, img, label, common_transforms=True, metadata=None, size=(256, 256), augmentation_dict=None,
                 dim_in=3, debug=False, cells_and_nuclei=False, target_segmentation="N", channel_invariant = False):
        self.X = img
        self.Y = label
        self.common_transforms = common_transforms

        assert len(self.X) == len(self.Y), "The number of images and labels must be the same"
        if len(metadata) == 0:
            self.metadata = [None] * len(self.X)
        else:
            self.metadata = metadata

        assert len(self.X) == len(self.metadata), print("The number of images and metadata must be the same")
        self.size = size
        self.Augmenter = Augmentations(augmentation_dict=augmentation_dict, debug=debug, shape=self.size,
                                       dim_in=dim_in, cells_and_nuclei=cells_and_nuclei,
                                       target_segmentation=target_segmentation, channel_invariant = channel_invariant)

    def __len__(self):
        return len(self.X)

    def __getitem__(self, i):

        data = self.X[i]
        label = self.Y[i]
        meta = self.metadata[i]

        if self.common_transforms:
            data, label = self.Augmenter(data, label, meta)

        if len(label.shape) == 2:
            label = label[None, :]
        if len(data.shape) == 2:
            data = data[None, :]

        assert not data.isnan().any(), "Tranformed images contains NaN"
        assert not label.isnan().any(), "Transformed labels contains NaN"

        return data.float(), label



def plot_loss(_model):
    loss_fig = plt.figure()
    timer = loss_fig.canvas.new_timer(interval=300000)
    timer.add_callback(plt.close)

    losses = [param.grad.norm().item() for name, param in _model.named_parameters() if param.grad is not None]
    names = [name for name, param in _model.named_parameters() if param.grad is not None]

    plt.plot(losses)
    plt.xticks(np.arange(len(names))[::1], names[::1])
    plt.xticks(fontsize=8, rotation=90)
    spacing = 0.5
    loss_fig.subplots_adjust(bottom=spacing)
    timer.start()
    plt.show()



def check_max_grad(_model):
    losses = np.array([param.grad.norm().item() for name, param in _model.named_parameters() if param.grad is not None])
    return losses.max()


def check_min_grad(_model):
    losses = np.array([param.grad.norm().item() for name, param in _model.named_parameters() if param.grad is not None])
    return losses.min()


def check_mean_grad(_model):
    losses = np.array([param.grad.norm().item() for name, param in _model.named_parameters() if param.grad is not None])
    return losses.mean()



def optimize_hyperparameters(model,postprocessing_fn, data_loader = None, val_images = None, val_labels = None,max_evals = 50, verbose = False, threshold = [0.5, 0.7, 0.9], show_progressbar = True, device = None):


    from instanseg.utils.metrics import _robust_average_precision
    from instanseg.utils.utils import _choose_device

    from hyperopt import fmin
    from hyperopt import hp
    from hyperopt import Trials
    from hyperopt import tpe
    import copy

    if device is None:
        device = _choose_device()

    bayes_trials = Trials()

    space = {  # instanseg
        'mask_threshold': hp.uniform('mask_threshold', 0.3, 0.7),
        'seed_threshold': hp.uniform('seed_threshold', 0.5, 1),
        'overlap_threshold': hp.uniform('overlap_threshold', 0.1, 0.9),
        #'min_size': hp.uniform('min_size', 0, 30),
        'peak_distance': hp.uniform('peak_distance', 3, 10),
        'mean_threshold': hp.uniform('mean_threshold', 0.0, 0.3)} #the max could be increased, but may cuase the method not to converge for some reason.
    
    _model = model # copy.deepcopy(model)
    _model.eval()
    predictions = []

    with torch.no_grad():
        if data_loader is not None:
            for image_batch, labels_batch, _ in data_loader:
                    image_batch = image_batch.to(device)
                    output = _model(image_batch).cpu()
                    predictions.extend([pred,masks] for pred,masks in zip(output,labels_batch))


            def objective(params={}):
                pred_masks = []
                gt_masks = []
                for pred, masks in predictions:
                    lab = postprocessing_fn(pred.to(device), **params).cpu()
                    pred_masks.append(lab)
                    gt_masks.append(masks)

                mean_f1 = _robust_average_precision(torch.stack(gt_masks),torch.stack(pred_masks),threshold = threshold)

                if type(mean_f1) == list:
                    mean_f1 = np.nanmean(mean_f1)

                return 1 - mean_f1
        
        elif val_images is not None and val_labels is not None:
            from instanseg.utils.tiling import _instanseg_padding, _recover_padding
            def objective(params={}):
                pred_masks = []
                gt_masks = []
                #randomly shuffle val_images and val_labels

                np.random.seed(0)
                indexes = np.random.permutation(len(val_images))[:300]
                indexes.sort()

                for i in indexes:
                    imgs = val_images[i]
                    gt_mask = val_labels[i]
                    with torch.no_grad():
                        imgs = imgs.to(device)
                        imgs, pad = _instanseg_padding(imgs, min_dim = 32)
                        output = _model(imgs[None,])
                        output = _recover_padding(output, pad).squeeze(0)
                        lab = postprocessing_fn(output.to(device), **params).cpu()
                        pred_masks.append(lab)
                        gt_masks.append(gt_mask)

                mean_f1 = _robust_average_precision(gt_masks,pred_masks,threshold = threshold)

                if type(mean_f1) == list:
                    mean_f1 = np.nanmean(mean_f1)

                return 1 - mean_f1
        else:
            raise ValueError("Either data_loader or val_images and val_labels must be provided")

        print("Optimizing hyperparameters")
        # Optimize
        best = fmin(fn=objective, space=space, algo=tpe.suggest,
                    max_evals=max_evals, trials=bayes_trials, show_progressbar = show_progressbar)
    
    if verbose:
        print(best)
    return best






import numpy as np


def build_monai_model(model_str: str, build_model_dictionary: dict):


    if model_str == "AttentionUNet":
        from monai.networks.nets import AttentionUnet

        model = AttentionUnet(spatial_dims=2, in_channels=int(build_model_dictionary["dim_in"]),
                              out_channels=build_model_dictionary["dim_out"], \
                              dropout=build_model_dictionary["dropprob"], channels=build_model_dictionary["layers"], \
                              strides=tuple([2 for _ in build_model_dictionary["layers"][:-1]])
                              )
    elif model_str == "FlexibleUNet":
        from monai.networks.nets import FlexibleUNet
        model = FlexibleUNet(in_channels=build_model_dictionary["dim_in"],
                             out_channels=build_model_dictionary["dim_out"], dropout=build_model_dictionary["dropprob"],
                             backbone="efficientnet-b0")
        

    elif model_str == "BasicUNetPlusPlus":
        from monai.networks.nets import BasicUNetPlusPlus
        model = BasicUNetPlusPlus(spatial_dims=2, in_channels=build_model_dictionary["dim_in"],
                                  out_channels=build_model_dictionary["dim_out"],
                                  dropout=build_model_dictionary["dropprob"])

        class ModelWrapper(BasicUNetPlusPlus):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def forward(self, inputs):
                output = super().forward(inputs)
                modified_output = output[0]  # Modify the output here as needed
                return modified_output

        model = ModelWrapper(spatial_dims=2, in_channels=build_model_dictionary["dim_in"],
                             out_channels=build_model_dictionary["dim_out"], dropout=build_model_dictionary["dropprob"])

    elif model_str == "UNETR":
        from monai.networks.nets import UNETR
        model = UNETR(in_channels=build_model_dictionary["dim_in"], out_channels=build_model_dictionary["dim_out"],
                      img_size=256, feature_size=32, norm_name='batch', spatial_dims=2)


    else:
        raise NotImplementedError("Model not implemented: " + model_str)

    return model


def read_model_args_from_csv(path=r"../results/", folder=""):
    import pandas as pd
    from pathlib import Path
    model_path = Path(path) / folder
    df = pd.read_csv(model_path / "experiment_log.csv", header=None)
    build_model_dictionary = dict(zip(list(df[0]), list(df[1])))

    if "model_shape" in build_model_dictionary.keys():
        build_model_dictionary["model_shape"] = eval(build_model_dictionary["model_shape"])
    for key in ["dim_in", "n_sigma", "dim_out", "dim_coords"]:
        build_model_dictionary[key] = eval(str(build_model_dictionary[key])) if str(
            build_model_dictionary[key]) != "nan" else None
    if "to_centre" in build_model_dictionary.keys():
        build_model_dictionary["to_centre"] = eval(build_model_dictionary["to_centre"])
    if "dropprob" in build_model_dictionary.keys():
        build_model_dictionary["dropprob"] = float(build_model_dictionary["dropprob"])
    if "layers" in build_model_dictionary.keys():
        build_model_dictionary["layers"] = tuple(eval(build_model_dictionary["layers"]))
    if "requested_pixel_size" in build_model_dictionary.keys():
        build_model_dictionary["pixel_size"] = float(build_model_dictionary["requested_pixel_size"])
    if "cells_and_nuclei" in build_model_dictionary.keys():
        build_model_dictionary["cells_and_nuclei"] = bool(eval(build_model_dictionary["cells_and_nuclei"]))
    if "norm" in build_model_dictionary.keys():
        if build_model_dictionary["norm"] == "None" or str(build_model_dictionary["norm"]).lower() == "nan":
            build_model_dictionary["norm"] = None
        else:
            build_model_dictionary["norm"] = str(build_model_dictionary["norm"])
    else:
        print("Norm not specified in model dictionary")
        build_model_dictionary["norm"] = None
    if "feature_engineering" in build_model_dictionary.keys():
        build_model_dictionary["feature_engineering"] = str(build_model_dictionary["feature_engineering"])
    else:
        print("Feature engineering not specified in model dictionary")
        build_model_dictionary["feature_engineering"] = "0"
    if "adaptor_net_str" in build_model_dictionary.keys():
        build_model_dictionary["adaptor_net_str"] = str(build_model_dictionary["adaptor_net_str"])
    if "multihead" in build_model_dictionary.keys():
        build_model_dictionary["multihead"] = bool(eval(build_model_dictionary["multihead"]))
    else:
        build_model_dictionary["multihead"] = False
    if "channel_invariant" in build_model_dictionary.keys():
        build_model_dictionary["channel_invariant"] = bool(eval(build_model_dictionary["channel_invariant"]))

    return build_model_dictionary


def build_model_from_dict(build_model_dictionary):
    if build_model_dictionary["dim_in"] == 0 or build_model_dictionary["dim_in"] is None:
        dim_in = 3  # Channel invariance currently outputs a 3 channel image
    else:
        dim_in = build_model_dictionary["dim_in"]

    if "dropprob" not in build_model_dictionary.keys():
        build_model_dictionary["dropprob"] = 0.0

    if build_model_dictionary["model_str"] == "InstanSeg_UNet":
            from instanseg.utils.models.InstanSeg_UNet import InstanSeg_UNet
            print("Generating InstanSeg_UNet")
            multihead = build_model_dictionary["multihead"]

            if build_model_dictionary["cells_and_nuclei"]:
                if not multihead:
                    from itertools import chain
                    out_channels = [[build_model_dictionary["dim_coords"], build_model_dictionary["n_sigma"],1] for i in range(2)]
                    out_channels = list(chain(*out_channels))
                
                else:
                    out_channels = [[build_model_dictionary["dim_coords"], build_model_dictionary["n_sigma"],1] for i in range(2)]

            else:
                if not multihead:
                    out_channels = [[build_model_dictionary["dim_coords"], build_model_dictionary["n_sigma"],1]]
                else:
                    out_channels = [[build_model_dictionary["dim_coords"]], [build_model_dictionary["n_sigma"]],[1]]

            model = InstanSeg_UNet(in_channels=dim_in, 
                            layers = np.array(build_model_dictionary["layers"])[::-1],
                            out_channels=out_channels,
                            norm  = build_model_dictionary["norm"], 
                            dropout=build_model_dictionary["dropprob"])
            
    else:
        model = build_monai_model(build_model_dictionary["model_str"], build_model_dictionary)

    return model


def remove_module_prefix_from_dict(dictionary):
    """
    Removes the module prefix from a dictionary of model weights
    :param dictionary: dictionary of model weights
    :return: modified dictionary
    """
    modified_dict = {}
    for key, value in dictionary.items():
        if key.startswith('module.'):
            modified_dict[key[7:]] = value
        else:
            modified_dict[key] = value
    return modified_dict


def has_pixel_classifier_state_dict(state_dict):
    return bool(sum(['pixel_classifier' in key for key in state_dict.keys()]))


def has_adaptor_net_state_dict(state_dict):
    return bool(sum(['AdaptorNet' in key for key in state_dict.keys()]))

def has_pixel_classifier_model(model):
    import torch
    for module in model.modules():
        if isinstance(module, torch.nn.Module):
            module_class = module.__class__.__name__
            if 'pixel_classifier' in module_class or 'ProbabilityNet' in module_class:
                return True
    return False


def load_model_weights(model, device, folder, path=r"../models/", dict = None):
    import torch
    from pathlib import Path
    model_path = Path(path) / folder
    if torch.cuda.is_available():
        model_dict = torch.load(model_path / "model_weights.pth", weights_only= False)
    else:
        if device is None:
            if torch.backends.mps.is_available():
                device = 'mps'
                print('CUDA not available - attempting to load MPS model')
            else:
                device = 'cpu'
                print('CUDA not available - attempting to load CPU model')
        model_dict = torch.load(model_path / "model_weights.pth", map_location=device)

    model_dict['model_state_dict'] = remove_module_prefix_from_dict(model_dict['model_state_dict'])

    if has_pixel_classifier_state_dict(model_dict['model_state_dict']) and not has_pixel_classifier_model(model):
        from instanseg.utils.loss.instanseg_loss import InstanSeg

        method = InstanSeg(n_sigma=int(dict["n_sigma"]), feature_engineering_function= dict["feature_engineering"],dim_coords = dict["dim_coords"],device =device)
        model = method.initialize_pixel_classifier(model, MLP_width=int(dict["mlp_width"]))

    from instanseg.utils.models.ChannelInvariantNet import AdaptorNetWrapper, has_AdaptorNet
    if has_adaptor_net_state_dict(model_dict['model_state_dict']) and not has_AdaptorNet(model):
        from instanseg.utils.models.ChannelInvariantNet import AdaptorNetWrapper, has_AdaptorNet
        model = AdaptorNetWrapper(model, norm = dict["norm"],adaptor_net_str = dict["adaptor_net_str"])

    #from instanseg.utils.AI_utils import set_running_stats
    #set_running_stats(model,device = "cuda")

    model.load_state_dict(model_dict['model_state_dict'], strict=True)
    model.to(device)

    return model, model_dict

def load_model(folder,path=r"../models/", device='cpu'):
    build_model_dictionary = read_model_args_from_csv(path=path, folder=folder)

    empty_model = build_model_from_dict(build_model_dictionary)

    model, _ = load_model_weights(empty_model, path=path, folder=folder, device=device, dict = build_model_dictionary)

    return model, build_model_dictionary




from __future__ import annotations
import torch
import torch.nn as nn
from einops import rearrange
import torch.nn.functional as F



def conv_norm_act(in_channels, out_channels, sz,norm, act = "ReLU", depthwise = False):

    if norm == "None" or norm is None:
        norm_layer = nn.Identity()
    elif norm.lower() == "batch":
        norm_layer = nn.BatchNorm2d(out_channels,eps = 1e-5, momentum = 0.05)
    elif norm.lower() == "instance":
        norm_layer = nn.InstanceNorm2d(out_channels,eps = 1e-5, track_running_stats=False, affine = True)
    else:
        raise ValueError("Norm must be None, batch or instance")
    
    if act == "None" or act is None:
        act_layer = nn.Identity()
    elif act.lower() == "relu":
        act_layer = nn.ReLU(inplace=True)
    elif act.lower() == "relu6":
        act_layer = nn.ReLU6(inplace=True)
    elif act.lower() == "mish":
        act_layer = nn.Mish(inplace=True)
    else:
        raise ValueError("Act must be None, ReLU or Mish")
    
    if depthwise:
        return nn.Sequential(
            nn.Conv2d(in_channels, out_channels, sz, padding=sz//2, groups = in_channels),
            norm_layer,
            act_layer,
        )
    else:
        
        return nn.Sequential(
            nn.Conv2d(in_channels, out_channels, sz, padding=sz//2),
            norm_layer,
            act_layer,
        )  


class DecoderBlock(nn.Module):
    def __init__(
            self,
            in_channels,
            skip_channels,
            out_channels,
            norm = "BATCH",
            act = "ReLU",
            shallow = False,
    ):
        super().__init__()

        
        self.conv0 = conv_norm_act(in_channels, out_channels, 1,norm,act)
        self.conv_skip = conv_norm_act(skip_channels, out_channels, 1,norm,act)
        self.conv1 = conv_norm_act(in_channels, out_channels, 3,norm,act)
        self.conv2 = conv_norm_act(out_channels, out_channels, 3,norm,act)
        self.conv3 = conv_norm_act(out_channels, out_channels, 3,norm,act)
        self.conv4 = conv_norm_act(out_channels, out_channels, 3,norm,act)

        if shallow:
            self.conv3 = nn.Identity()

    def forward(self, x, skip=None):
        x = F.interpolate(x, scale_factor=2, mode="nearest")
        proj = self.conv0(x)
        x = self.conv1(x)
        x =  proj + self.conv2(x + self.conv_skip(skip))
        x = x + self.conv4(self.conv3(x))
        return x


class EncoderBlock(nn.Module):
    def __init__(
            self,
            in_channels,
            out_channels,
            pool = True,
            norm = "BATCH",
            act = "ReLU",
            shallow = False,
    ):
        super().__init__()

        if pool:
            self.maxpool = nn.MaxPool2d(2, 2)
        else:
            self.maxpool = nn.Identity()
        self.conv0 = conv_norm_act(in_channels, out_channels, 1,norm,act)
        self.conv1 = conv_norm_act(in_channels, out_channels, 3,norm,act)
        self.conv2 = conv_norm_act(out_channels, out_channels, 3,norm,act)
        self.conv3 = conv_norm_act(out_channels, out_channels, 3,norm,act)
        self.conv4 = conv_norm_act(out_channels, out_channels, 3,norm,act)

        if shallow:
            self.conv2 = nn.Identity()
            self.conv3 = nn.Identity()

    def forward(self, x):

        x = self.maxpool(x)
        proj = self.conv0(x)
        x = self.conv1(x)
        x =  proj + self.conv2(x)
        x = x + self.conv4(self.conv3(x))
        return x
    
class Decoder(nn.Module):
    def __init__(
            self,
            layers,
            out_channels,
            norm,
            act):
        super().__init__()
        
        self.decoder = nn.ModuleList([DecoderBlock(layers[i],layers[i+1],layers[i+1],norm = norm, act =  act) for i in range(len(layers)-1)])
        
        self.final_block = nn.ModuleList([conv_norm_act(layers[-1],out_channel,1, norm = norm if (norm is not None) and norm.lower() != "instance" else None,act = None) for out_channel in out_channels])
        
    def forward(self,x,skips):
        for layer,skip in zip(self.decoder,skips[::-1]):
            x = layer(x,skip)

        x = torch.cat([final_block(x) for final_block in self.final_block],dim = 1)
        return x
            
    
class InstanSeg_UNet(nn.Module):
    def __init__(self,in_channels,out_channels,layers = [256,128,64,32],norm = "BATCH",dropout = 0, act = "ReLu"):
        super().__init__()
        layers = layers[::-1]
        self.encoder = nn.ModuleList([EncoderBlock(in_channels,layers[0],pool = False,norm = norm, act = act)] + [EncoderBlock(layers[i],layers[i+1],norm = norm, act = act) for i in range(len(layers)-1)])
        layers = layers[::-1]

        # out_channels should be a list of lists [[2,2,1],[2,2,1]] means two decoders, each with 3 output blocks. The output will be of shape 10.

        if type(out_channels) == int:
            out_channels = [[out_channels]]
        if type(out_channels[0]) == int:
            out_channels = [out_channels]
            
        self.decoders = nn.ModuleList([Decoder(layers,out_channel,norm, act) for out_channel in out_channels])
    
    def forward(self,x):
        skips = []
        for n,layer in enumerate(self.encoder):
            x = layer(x)
            if n < len(self.encoder)-1:
                skips.append(x)

        return torch.cat([decoder(x,skips) for decoder in self.decoders],dim = 1)
    



if __name__ == "__main__":

    net = InstanSeg_UNet(
        in_channels=3,
        out_channels=[5,3],
    )

    print(net(torch.randn(1,3,256,256)).shape)



from typing import Union, List, Optional, Tuple
import numpy as np
import torch
from torch import nn
from torch.nn.functional import interpolate
from pathlib import Path, PosixPath
from tiffslide import TiffSlide
import zarr
import os
from instanseg.utils.pytorch_utils import _to_tensor_float32, _to_ndim



class InstanSeg():
    """
    Main class for running InstanSeg.
    """
    def __init__(self, 
                 model_type: Union[str,nn.Module] = "brightfield_nuclei", 
                 device: Optional[str] = None, 
                 image_reader: str = "tiffslide",
                 verbosity: int = 1 #0,1,2
                 ):
        
        """
        :param model_type: The type of model to use. If a string is provided, the model will be downloaded. If the model is not public, it will look for a model in your bioimageio folder. If an nn.Module is provided, this model will be used.
        :param device: The device to run the model on. If None, the device will be chosen automatically.
        :param image_reader: The image reader to use. Options are "tiffslide", "skimage.io", "bioio", "AICSImageIO".
        :param verbosity: The verbosity level. 0 is silent, 1 is normal, 2 is verbose.
        """
        from instanseg.utils.utils import download_model, _choose_device

        self.verbosity = verbosity
        self.verbose = verbosity != 0

        if isinstance(model_type, nn.Module):
            self.instanseg = model_type
        else:
            self.instanseg = download_model(model_type, verbose = self.verbose)
        self.inference_device = _choose_device(device, verbose= self.verbose)
        self.instanseg = self.instanseg.to(self.inference_device)

        self.prefered_image_reader = image_reader
        self.small_image_threshold = 3 * 1500 * 1500 #max number of image pixels to be processed on GPU.
        self.medium_image_threshold = 10000 * 10000 #max number of image pixels that could be loaded in RAM.
        self.prediction_tag = "_instanseg_prediction"

    def read_image(self, image_str: str) -> Union[Tuple[str, float], Tuple[np.ndarray, float]]:
        """
        Read an image file from disk.
        :param image_str: The path to the image.
        :return: The image array if it can be safely read (or the path to the image if it cannot) and the pixel size in microns.
        """
        if self.prefered_image_reader == "tiffslide":
            from tiffslide import TiffSlide
            slide = TiffSlide(image_str)
            img_pixel_size = slide.properties['tiffslide.mpp-x']
            width,height = slide.dimensions[0], slide.dimensions[1]
            num_pixels = width * height
            if num_pixels < self.medium_image_threshold:
                image_array = slide.read_region((0, 0), 0, (width, height), as_array=True)
            else:
                return image_str, img_pixel_size
            
        elif self.prefered_image_reader == "skimage.io":
            from skimage.io import imread
            image_array = imread(image_str)
            img_pixel_size = None

        elif self.prefered_image_reader == "bioio":
            from bioio import BioImage
            slide = BioImage(image_str)
            img_pixel_size = slide.physical_pixel_sizes.X
            num_pixels = np.cumprod(slide.shape)[-1]
            if num_pixels < self.medium_image_threshold:
                image_array = slide.get_image_data().squeeze()
            else:
                return image_str, img_pixel_size
        else:
            raise NotImplementedError(f"Image reader {self.prefered_image_reader} is not implemented.")
        
        if img_pixel_size is None or float(img_pixel_size) < 0 or float(img_pixel_size) > 2:
            img_pixel_size = self.read_pixel_size(image_str)

        if img_pixel_size is not None:
            import warnings
            if float(img_pixel_size) <= 0 or float(img_pixel_size) > 2:
                warnings.warn(f"Pixel size {img_pixel_size} microns per pixel is invalid.")
                img_pixel_size = None

        return image_array, img_pixel_size
    
    def read_pixel_size(self,image_str: str) -> float:
        """
        Read the pixel size from an image on disk.
        :param image_str: The path to the image.
        :return: The pixel size in microns.
        """
        try:
            from tiffslide import TiffSlide
            slide = TiffSlide(image_str)
            img_pixel_size = slide.properties['tiffslide.mpp-x']
            if img_pixel_size is not None and img_pixel_size > 0 and img_pixel_size < 2:
                return img_pixel_size
        except Exception as e:
            print(e)
            pass
        from bioio import BioImage
        try:
            slide = BioImage(image_str)
            img_pixel_size = slide.physical_pixel_sizes.X
            if img_pixel_size is not None and img_pixel_size > 0 and img_pixel_size < 2:
                return img_pixel_size
        except Exception as e:
            print(e)
            pass
        import slideio
        try:
            slide = slideio.open_slide(image_str, driver = "AUTO")
            scene  = slide.get_scene(0)
            img_pixel_size = scene.resolution[0] * 10**6

            if img_pixel_size is not None and img_pixel_size > 0 and img_pixel_size < 2:
                    
                return img_pixel_size
        except Exception as e:
            print(e)
            pass
        print("Could not read pixel size from image metadata.")
        
        return None


    def read_slide(self, image_str: str):
        """
        Read a whole slide image from disk.
        :param image_str: The path to the image.
        """
        if self.prefered_image_reader == "tiffslide":
            slide = TiffSlide(image_str)
        # elif self.prefered_image_reader == "AICSImageIO":
        #     from aicsimageio import AICSImage
        #     slide = AICSImage(image_str)
        # elif self.prefered_image_reader == "bioio":
        #     from bioio import BioImage
        #     slide = BioImage(image_str)
        # elif self.prefered_image_reader == "slideio":
        #     import slideio
        #     slide = slideio.open_slide(image_str, driver = "AUTO")

        else:
            raise NotImplementedError(f"Image reader {self.prefered_image_reader} is not implemented for whole slide images.")
        return slide
    
    def _to_tensor(self, image: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        return _to_tensor_float32(image)
    
    def _normalise(self, image: torch.Tensor) -> torch.Tensor:
        from instanseg.utils.utils import percentile_normalize, _move_channel_axis
        assert image.ndim == 3 or image.ndim == 4, f"Input image shape {image.shape} is not supported."
        if image.dim() == 3:
            image = percentile_normalize(image)
            image = image[None]
        else:
            image = torch.stack([percentile_normalize(i) for i in image])

        return image

    def eval(self,
             image: Union[str, List[str]], 
             pixel_size: Optional[float] = None,
             save_output: bool = False,
             save_overlay: bool = False,
             save_geojson: bool = False,
             **kwargs) -> Union[torch.Tensor, List[torch.Tensor], None]:
        """
        Evaluate the input image or list of images using the InstanSeg model.
        :param image: The path to the image, or a list of such paths.
        :param pixel_size: The pixel size in microns.
        :param save_output: Controls whether the output is saved to disk (see :func:`save_output <instanseg.Instanseg.save_output>`).
        :param save_overlay: Controls whether the output is saved to disk as an overlay (see :func:`save_output <instanseg.Instanseg.save_output>`).
        :param save_geojson: Controls whether the geojson output labels are saved to disk (see :func:`save_output <instanseg.Instanseg.save_output>`).
        :param kwargs: Passed to other eval methods, eg :func:`save_output <instanseg.Instanseg.eval_small_image>`, :func:`save_output <instanseg.Instanseg.eval_medium_image>`, :func:`save_output <instanseg.Instanseg.eval_whole_slide_image>` 
        :return: A torch.Tensor of outputs if the input is a path to a single image, or a list of such outputs if the input is a list of paths, or None if the input is a whole slide image.
        """

        if isinstance(image, PosixPath):
            image = str(image)
        if isinstance(image, str):
            initial_type = "not_list"
            image_list = [image]
        else:
            initial_type = "list"
            image_list = image
        
        output_list = []
    
        for image in image_list:
            image_array, img_pixel_size = self.read_image(image)

            if img_pixel_size is None and pixel_size is not None:
                img_pixel_size = pixel_size
            if img_pixel_size is None:
                import warnings
                warnings.warn("Pixel size not provided and could not be read from image metadata, this may lead to innacurate results.")
                
            if not isinstance(image_array, str):
                
                num_pixels = np.cumprod(image_array.shape)[-1]
                if num_pixels < self.small_image_threshold:
                    instances = self.eval_small_image(image = image_array, 
                                                       pixel_size = img_pixel_size, 
                                                       return_image_tensor=False, **kwargs)

                else:
                    instances = self.eval_medium_image(image = image_array, 
                                                       pixel_size = img_pixel_size, 
                                                       return_image_tensor=False, **kwargs)

                output_list.append(instances)

                if save_output:
                    self.save_output(image, instances, image_array = image_array, save_overlay = save_overlay, save_geojson = save_geojson)
     
                    
            else:
                self.eval_whole_slide_image(image_array, pixel_size, save_geojson = save_geojson, **kwargs)
                output_list.append(None)

        if initial_type == "not_list":
            output = output_list[0]
        else:
            output = output_list
        
        return output
    
    def save_output(self,
                    image_path: str, 
                    labels: torch.Tensor,
                    image_array: Optional[np.ndarray] = None,
                    save_overlay = False,
                    save_geojson = False) -> None:
        """
        Save the output of InstanSeg to disk.
        :param image_path: The path to the image, and where outputs will be saved.
        :param labels: The output labels.
        :param image_array: The image in array format. Required to save overlay.
        :param save_overlay: Save the labels overlaid on the image.
        :param save_geojson: Save the labels as a GeoJSON feature collection.
        """
        import os

        if isinstance(image_path, str):
            image_path = Path(image_path)
        if isinstance(labels, torch.Tensor):
            labels = labels.cpu().detach().numpy()

        new_stem = image_path.stem + self.prediction_tag

        from skimage import io

        if self.verbose:

            out_path = Path(image_path).parent / (new_stem + ".tiff")
            print(f"Saving output to {out_path}")
            io.imsave(out_path, labels.squeeze().astype(np.int32), check_contrast=False)

        if save_geojson:

            labels = _to_ndim(labels, 4)
        
            output_dimension = labels.shape[1]
            from instanseg.utils.utils import labels_to_features
            import json
            if output_dimension == 1:
                features = labels_to_features(labels[0,0],object_type = "detection")

            elif output_dimension == 2:
             #   breakpoint()
                features = labels_to_features(labels[0,0],object_type = "detection",classification="Nuclei")["features"] + labels_to_features(labels[0,1],object_type = "detection",classification = "Cells")["features"]
            geojson = json.dumps(features)

            geojson_path = Path(image_path).parent / (new_stem + ".geojson")
            with open(os.path.join(geojson_path), "w") as outfile:
                outfile.write(geojson)
        
        if save_overlay:

            if self.verbose:
                out_path = Path(image_path).parent / (new_stem + "_overlay.tiff")
                print(f"Saving overlay to {out_path}")
            assert image_array is not None, "Image array must be provided to save overlay."
            display = self.display(image_array, labels)
            
            io.imsave(out_path, display, check_contrast=False)



    def eval_small_image(self,
                         image: torch.Tensor,
                         pixel_size: Optional[float] = None,
                         normalise: bool = True,
                         return_image_tensor: bool = True,
                         target: str = "all_outputs", #or "nuclei" or "cells"
                         rescale_output: bool = True,
                         **kwargs) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Evaluate a small input image using the InstanSeg model.
        
        :param image:: The input image(s) to be evaluated.
        :param pixel_size: The pixel size of the image, in microns. If not provided, it will be read from the image metadata.
        :param normalise: Controls whether the image is normalised.
        :param return_image_tensor: Controls whether the input image is returned as part of the output.
        :param target: Controls what type of output is given, usually "all_outputs", "nuclei", or "cells".
        :param rescale_output: Controls whether the outputs should be rescaled to the same coordinate space as the input (useful if the pixel size is different to that of the InstanSeg model being used).
        :param kwargs: Passed to pytorch.
        
        :return: A tensor corresponding to the output targets specified, as well as the input image if requested.
        """
        from instanseg.utils.utils import percentile_normalize, _filter_kwargs

        image = _to_tensor_float32(image)

        image = _to_ndim(image, 4)

        original_shape = image.shape

        if pixel_size is not None:
            image = _rescale_to_pixel_size(image, pixel_size, self.instanseg.pixel_size)

            if original_shape[-2] != image.shape[-2] or original_shape[-1] != image.shape[-1]:
                img_has_been_rescaled = True
            else:
                img_has_been_rescaled = False

        image = image.to(self.inference_device)

        assert image.dim() ==3 or image.dim() == 4, f"Input image shape {image.shape} is not supported."

        if normalise:
                image = _to_ndim(image, 4)
                image = torch.stack([percentile_normalize(i) for i in image]) #over the batch dimension

        if target != "all_outputs" and self.instanseg.cells_and_nuclei:
            assert target in ["nuclei", "cells"], "Target must be 'nuclei', 'cells' or 'all_outputs'."
            if target == "nuclei":
                target_segmentation = torch.tensor([1,0])
            else:
                target_segmentation = torch.tensor([0,1])
        else:
            target_segmentation = torch.tensor([1,1])

        with torch.amp.autocast('cuda'):
            instanseg_kwargs = _filter_kwargs(self.instanseg, kwargs)
            instances = self.instanseg(image,target_segmentation = target_segmentation, **instanseg_kwargs)

        if pixel_size is not None and img_has_been_rescaled and rescale_output:  
            instances = interpolate(instances, size=original_shape[-2:], mode="nearest")

            if return_image_tensor:
                image = interpolate(image, size=original_shape[-2:], mode="bilinear")

        if return_image_tensor:
            return instances.cpu(), image.cpu()
        else:
            return instances.cpu()

    def eval_medium_image(self,
                          image: torch.Tensor, 
                          pixel_size: Optional[float] = None, 
                          normalise: bool = True,
                          tile_size: int = 512,
                          batch_size: int = 1,
                          return_image_tensor: bool = True,
                          normalisation_subsampling_factor: int = 1,
                          target: str = "all_outputs", #or "nuclei" or "cells"
                          rescale_output: bool = True,
                          **kwargs) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Evaluate a medium input image using the InstanSeg model. The image will be split into tiles, and then inference and object merging will be handled internally.
        
        :param image:: The input image(s) to be evaluated.
        :param pixel_size: The pixel size of the image, in microns. If not provided, it will be read from the image metadata.
        :param normalise: Controls whether the image is normalised.
        :param tile_size: The width/height of the tiles that the image will be split into.
        :param batch_size: The number of tiles to be run simultaneously.
        :param return_image_tensor: Controls whether the input image is returned as part of the output.
        :param normalisation_subsampling_factor: The subsampling or downsample factor at which to calculate normalisation parameters.
        :param target: Controls what type of output is given, usually "all_outputs", "nuclei", or "cells".
        :param rescale_output: Controls whether the outputs should be rescaled to the same coordinate space as the input (useful if the pixel size is different to that of the InstanSeg model being used).
        :param kwargs: Passed to pytorch.
        
        :return: A tensor corresponding to the output targets specified, as well as the input image if requested.
        """

        from instanseg.utils.utils import percentile_normalize, _move_channel_axis, _filter_kwargs

    
        image = _to_tensor_float32(image)
        
        from instanseg.utils.tiling import _sliding_window_inference
        original_shape = image.shape
        original_ndim = image.dim()

        if pixel_size is None:
            import warnings
            warnings.warn("Pixel size not provided, this may lead to innacurate results.")
        else:
            image = _rescale_to_pixel_size(image, pixel_size, self.instanseg.pixel_size)

            if original_shape[-2] != image.shape[-2] or original_shape[-1] != image.shape[-1]:
                img_has_been_rescaled = True
            else:
                img_has_been_rescaled = False
        

        image = _to_ndim(image, 3)

       # breakpoint()
        
        if normalise:
            image = percentile_normalize(image, subsampling_factor=normalisation_subsampling_factor)
            
        output_dimension = 2 if self.instanseg.cells_and_nuclei else 1

        if target != "all_outputs" and output_dimension == 2:
            assert target in ["nuclei", "cells"], "Target must be 'nuclei', 'cells' or 'all_outputs'."
            if target == "nuclei":
                target_segmentation = torch.tensor([1,0])
            else:
                target_segmentation = torch.tensor([0,1])
            output_dimension = 1
        else:
            target_segmentation = torch.tensor([1,1])

        instanseg_kwargs = _filter_kwargs(self.instanseg, kwargs)

        instances = _sliding_window_inference(image,
                                              self.instanseg,
                                              window_size = (tile_size,tile_size),sw_device = self.inference_device,
                                              device = 'cpu', 
                                              batch_size= batch_size,
                                              output_channels = output_dimension,
                                              show_progress= self.verbose,
                                              target_segmentation = target_segmentation,
                                              **instanseg_kwargs).float()

        instances = _to_ndim(instances, 4)
        image = _to_ndim(image, 4)
        
        if pixel_size is not None and img_has_been_rescaled and rescale_output:  
            instances = interpolate(instances, size=original_shape[-2:], mode="nearest")
            instances = _to_ndim(instances, 4)

            if return_image_tensor:
                image = interpolate(image, size=original_shape[-2:], mode="bilinear")

        image = _to_ndim(image, original_ndim)

        if return_image_tensor:
            return instances.cpu(), image.cpu()
        else:
            return instances.cpu()
        
    def eval_whole_slide_image(self,
                               image: str,
                               pixel_size: Optional[float] = None, 
                               normalise: bool = True,
                               normalisation_subsampling_factor: int = 10,
                               tile_size: int = 512,
                               overlap: int = 100,
                               detection_size: int = 20, 
                               batch_size: int = 1,
                               save_geojson: bool = False,
                               use_otsu_threshold: bool = False,
                               **kwargs):
            """
            Evaluate a whole slide input image using the InstanSeg model. This function uses slideio to read an image and then segments it using the instanseg model. The segmentation is done in a tiled manner to avoid memory issues. 
            
            :param image: The input image to be evaluated.
            :param pixel_size: The pixel size of the image, in microns. If not provided, it will be read from the image metadata.
            :param normalise: Controls whether the image is normalised.
            :param tile_size: The width/height of the tiles that the image will be split into.
            :param overlap: The overlap (in pixels) betwene tiles.
            :param detection_size: The expected maximum size of detection objects.
            :param batch_size: The number of tiles to be run simultaneously.
            :param normalisation_subsampling_factor: The subsampling or downsample factor at which to calculate normalisation parameters.
            :param use_otsu_threshold: bool = False. Whether to use an otsu threshold on the image thumbnail to find the tissue region.
            :param kwargs: Passed to pytorch.
            :return: Returns a zarr file with the segmentation. The zarr file is saved in the same directory as the image with the same name but with the extension .zarr.
            """

            memory_block_size = tile_size,tile_size
            inference_tile_size = (tile_size,tile_size)

            from itertools import product
            from instanseg.utils.pytorch_utils import torch_fastremap, match_labels
            from pathlib import Path
            from tqdm import tqdm
            from instanseg.utils.tiling import _chops, _remove_edge_labels, _zarr_to_json_export
            from instanseg.utils.utils import show_images
            
            instanseg = self.instanseg

            image, img_pixel_size = self.read_image(image)
            slide = self.read_slide(image)

            n_dim = 2 if instanseg.cells_and_nuclei else 1
            model_pixel_size = instanseg.pixel_size

            new_stem = Path(image).stem + self.prediction_tag
            file_with_zarr_extension = Path(image).parent / (new_stem + ".zarr")

            if img_pixel_size > 1 or img_pixel_size < 0.1:
                import warnings
                warnings.warn("The image pixel size {} is not in microns.".format(img_pixel_size))
                if pixel_size is not None:
                    img_pixel_size = pixel_size
                else:
                    raise ValueError("The image pixel size {} is not in microns.".format(img_pixel_size))
            
                
            scale_factor = model_pixel_size/img_pixel_size

            dims = slide.dimensions
            dims = (int(dims[1]/ scale_factor), int(dims[0]/scale_factor)) #The dimensions are opposite to numpy/torch/zarr dimensions.

            pad2 = overlap + detection_size
            pad = overlap

            shape = memory_block_size
            chop_list = _chops(dims, shape, overlap=2*pad2)

            if use_otsu_threshold:
                mask,_ = _threshold_thumbnail(slide)
                valid_positions = _find_non_empty_positions(mask, chop_list, shape[0], dims)
            else:
                valid_positions = np.ones((len(chop_list[0])* len(chop_list[1])))

            chunk_shape = (n_dim,shape[0],shape[1])
            store = zarr.DirectoryStore(file_with_zarr_extension) 
            canvas = zarr.zeros((n_dim,dims[0],dims[1]), chunks=chunk_shape, dtype=np.int32, store=store, overwrite = True)

            running_max = 0

            total = len(chop_list[0]) * len(chop_list[1])
            counter = -1
            for _, ((i, window_i), (j, window_j)) in tqdm(enumerate(product(enumerate(chop_list[0]), enumerate(chop_list[1]))), total=total, colour = "green", desc = "Slide progress: "):
                counter += 1
                if valid_positions[counter] == 0:
                    continue

            #  input_data = scene.read_block((int(window_j*scale_factor), int(window_i*scale_factor), int(shape[0]*scale_factor), int(shape[1]*scale_factor)), size = shape)

                best_level = slide.get_best_level_for_downsample(scale_factor)
                downsample_factor = slide.level_downsamples[best_level]

                initial_pixel_size = img_pixel_size
                itermediate_pixel_size = initial_pixel_size * downsample_factor
                final_pixel_size = model_pixel_size

                intermediate_to_final = final_pixel_size/itermediate_pixel_size

                if np.allclose(intermediate_to_final,1, 0.01): #if you change this value, you MUST modify the _rescale_to_pixel_size function.
                    intermediate_to_final = 1
                
                # Calculate the size of the region needed at the base level to get the desired output size
                intermediate_shape = (int(shape[0] * intermediate_to_final), int(shape[1] * intermediate_to_final))

                input_data = slide.read_region((int(window_j*scale_factor), int(window_i*scale_factor)), best_level, (int(intermediate_shape[0]) , int(intermediate_shape[1])), as_array=True)
            
                input_tensor = self._to_tensor(input_data)

            
                new_tile = self.eval_small_image(input_tensor,
                                                  pixel_size = itermediate_pixel_size,
                                                  tile_size = inference_tile_size[0],
                                                  batch_size = batch_size,
                                                  return_image_tensor = False,
                                                  normalise = normalise,
                                                  normalisation_subsampling_factor = normalisation_subsampling_factor,
                                                    **kwargs)
                                                

                if not np.allclose(intermediate_to_final,1, 0.01):
                    from torch.nn.functional import interpolate
                    new_tile = interpolate(new_tile, size=shape[-2:], mode="nearest").int()[0]
                
                new_tile = _to_ndim(new_tile, 3)

                num_iter = new_tile.shape[0]

                for n in range(num_iter):
                    
                    ignore_list = []
                    if i == 0:
                        ignore_list.append("top")
                    if j == 0:
                        ignore_list.append("left")
                    if i == len(chop_list[0])-1:
                        ignore_list.append("bottom")
                    if j == len(chop_list[1])-1:
                        ignore_list.append("right")

                    if i == len(chop_list[0])-1 and j == len(chop_list[1])-1:
                        tile1 = canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]]
                        tile2 = _remove_edge_labels(new_tile[n,pad:shape[0],pad: shape[1]], ignore = ignore_list)

                    elif i == len(chop_list[0])-1:
                        tile1 = canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]]
                        tile2 = _remove_edge_labels(new_tile[n,pad:shape[0],pad : shape[1]], ignore =ignore_list)

                    elif j == len(chop_list[1])-1:
                        tile1 = canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]]
                        tile2 = _remove_edge_labels(new_tile[n,pad:shape[0],pad: shape[1]], ignore = ignore_list)

                    elif i == 0 and j == 0:
                        tile1 = canvas[n, window_i  :window_i + shape[0], window_j :window_j + shape[1]]
                        tile2 = _remove_edge_labels(new_tile[n, :shape[0], : shape[1]], ignore = ignore_list)
                    elif i == 0:
                        tile1 = canvas[n, window_i  :window_i + shape[0], window_j + pad :window_j + shape[1]]
                        tile2 = _remove_edge_labels(new_tile[n, :shape[0],pad : shape[1]], ignore = ignore_list)

                    elif j == 0:
                        tile1 = canvas[n, window_i  + pad:window_i + shape[0], window_j:window_j + shape[1]]
                        tile2 = _remove_edge_labels(new_tile[n,pad :shape[0],: shape[1]], ignore = ignore_list)

                    if j == 0 or i == 0 or j == len(chop_list[1])-1 or i == len(chop_list[0])-1:

                        tile2 = torch_fastremap(tile2)
                        tile2[tile2>0] = tile2[tile2>0] + running_max

                        tile1_torch = torch.tensor(np.array(tile1), dtype = torch.int32)

                        remapped = match_labels(tile1_torch, tile2, threshold = 0.1)[1]
                        tile1_torch[remapped>0] = remapped[remapped>0].int()

                        running_max = max(running_max, tile1_torch.max())

                        if i == len(chop_list[0])-1 and j == len(chop_list[1])-1:
                            canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                        elif i == len(chop_list[0])-1:
                            canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                        elif j == len(chop_list[1])-1:
                            canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                        elif i == 0 and j == 0:
                            canvas[n, window_i  :window_i + shape[0], window_j :window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                        elif i == 0:
                            canvas[n, window_i  :window_i + shape[0], window_j + pad :window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                        elif j == 0:
                            canvas[n, window_i  + pad:window_i + shape[0], window_j:window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)

                    else:
                        
                        tile1 = canvas[n, window_i + pad:window_i + shape[0] - pad, window_j + pad:window_j + shape[1] - pad]
                        tile2 = _remove_edge_labels(new_tile[n,pad:shape[0] -pad,pad: shape[1]-pad])
                        
                        tile2 = torch_fastremap(tile2)

                        tile2[tile2>0] = tile2[tile2>0] + running_max

                        tile1_torch = torch.tensor(np.array(tile1), dtype = torch.int32)
                        remapped = match_labels(tile1_torch, tile2, threshold = 0.1)[1]

                        tile1_torch[remapped>0] = remapped[remapped>0].int()
                        running_max = max(running_max, tile1_torch.max())
                    
                        canvas[n, window_i + pad:window_i + shape[0] - pad, window_j + pad:window_j + shape[1] - pad] = tile1_torch.numpy().astype(np.int32)

            if save_geojson:
                print("Exporting to geojson")
                _zarr_to_json_export(file_with_zarr_extension, detection_size = detection_size, size = shape[0], scale = scale_factor, n_dim = n_dim)
                    
    
    def display(self,
                image: torch.tensor,
                instances: torch.Tensor) -> np.ndarray:
        """
        Save the output of an InstanSeg model overlaid on the input.
        See :func:`save_image_with_label_overlay <instanseg.utils.save_image_with_label_overlay>` for more details and return types.
        :param image: The input image.
        :param instances: The output labels.
        """
        from instanseg.utils.utils import save_image_with_label_overlay

        if isinstance(image, torch.Tensor):
            image = image.cpu().detach().numpy()

        im_for_display = _display_colourized(image.squeeze())

        output_dimension = instances.shape[1]

        if output_dimension ==1: #Nucleus or cell mask
            labels_for_display = instances[0,0] #Shape is 1,H,W
            image_overlay = save_image_with_label_overlay(im_for_display,lab=labels_for_display,return_image=True, label_boundary_mode="thick", label_colors=None,thickness=10,alpha=0.9)
        elif output_dimension ==2: #Nucleus and cell mask
            nuclei_labels_for_display = instances[0,0]
            cell_labels_for_display = instances[0,1] #Shape is 1,H,W
            image_overlay = save_image_with_label_overlay(im_for_display,lab=nuclei_labels_for_display,return_image=True, label_boundary_mode="thick", label_colors="red",thickness=10)
            image_overlay = save_image_with_label_overlay(image_overlay,lab=cell_labels_for_display,return_image=True, label_boundary_mode="inner", label_colors="green",thickness=1)

        return image_overlay
    

    def _cluster_instances_by_mean_channel_intensity(self, image_tensor: torch.Tensor, labeled_output: torch.Tensor):

        #This is experimental code that is not yet implemented. You'll need to install rapids_singlecell, cuml and scanpy to run this code.

        from instanseg.utils.biological_utils import get_mean_object_features
        import fastremap
        import numpy as np
        from instanseg.utils.utils import apply_cmap
        from instanseg.utils.pytorch_utils import torch_fastremap
        import rapids_singlecell as rsc
        import scanpy as sc
        import matplotlib.pyplot as plt

        labeled_output = _to_ndim(labeled_output, 4)
        image_tensor = _to_ndim(image_tensor, 3)

        X_features = get_mean_object_features( image_tensor.to("cuda"), labeled_output.to("cuda"),)

        adata = sc.AnnData(X_features.cpu().numpy())
        rsc.get.anndata_to_GPU(adata)
        rsc.pp.scale(adata)
        rsc.pp.neighbors(adata, n_neighbors=8, n_pcs=50)
        rsc.tl.umap(adata)
        rsc.tl.leiden(adata, resolution=0.1)

        # Create the UMAP plot
        fig, axes = plt.subplots(1, 2, figsize=(15, 7))
        mapping = fastremap.component_map(np.arange(1, len(adata.obs["leiden"]) + 1), adata.obs["leiden"].astype(np.int64) + 1)
        labs = torch_fastremap(labeled_output[0, 0])
        labels = fastremap.remap(labs.numpy(), mapping, preserve_missing_labels=True)

        labels_disp = apply_cmap(labels, labels > 0, cmap="tab10")

        # Show the labeled image
        axes[0].imshow(labels_disp)
        axes[0].set_title('Labeled Image')
        axes[0].axis('off')

        sc.pl.umap(adata, color="leiden", legend_loc='on data', cmap="tab10", title='UMAP with Leiden Clustering', s=30, ax=axes[1], show = False)
        axes[1].axis('off')
        plt.subplots_adjust(wspace=0., hspace=0)
        plt.show()



def _threshold_thumbnail(slide, level=None, sigma = 3):
    from skimage.color import rgb2gray
    from skimage import filters
    import numpy as np

    if level is None:
        level = slide.level_count - 1

    img_thumbnail = slide.read_region((0, 0), level, size=(10000, 10000), as_array=True, padding=False)
    downsample_factor_thumbnail = slide.level_downsamples[level]

    gray_image = rgb2gray(np.array(img_thumbnail))
    threshold_value = filters.threshold_otsu(gray_image)
    gray_image = filters.gaussian(gray_image,sigma = sigma)>threshold_value
    binary_image = ~(gray_image > threshold_value)  # Apply the threshold to create a binary image

    return binary_image, downsample_factor_thumbnail



def _find_non_empty_positions(mask, chop_list, tile_size, chopped_image_size, emptiness_threshold = 0.1):
    """
    Precompute all valid positions within the mask where tiles can be placed.
    """
    from itertools import product
    from instanseg.utils.utils import show_images
    valid_positions = []

    downsample_factor_mask = chopped_image_size[0] / mask.shape[0]
    scaled_tile_size = int(round(tile_size / downsample_factor_mask,0))

    for y,x in product((chop_list[0]),(chop_list[1])):

        y = int(round(y / downsample_factor_mask,0))
        x = int(round(x / downsample_factor_mask,0))

        if mask[y:y + scaled_tile_size, x:x + scaled_tile_size].max() > emptiness_threshold:
            valid_positions.append(1)
        else:
            valid_positions.append(0)

    return valid_positions


def _rescale_to_pixel_size(image: torch.Tensor, 
                           requested_pixel_size: float, 
                           model_pixel_size: float) -> torch.Tensor:
    
    original_dim = image.dim()

    image = _to_ndim(image, 4)

    scale_factor = requested_pixel_size / model_pixel_size

    if not np.allclose(scale_factor,1, 0.01): #if you change this value, you MUST modify the whole_slide_image function.
        image = interpolate(image, scale_factor=scale_factor, mode="bilinear")

    return _to_ndim(image, original_dim)
    

def _display_colourized(mIF):
    from instanseg.utils.utils import _move_channel_axis, generate_colors

    mIF = _to_tensor_float32(mIF)
    mIF = mIF / (mIF.max() + 1e-6)
    if mIF.shape[0]!=3:
        colours = generate_colors(num_colors=mIF.shape[0])
        colour_render = (mIF.flatten(1).T @ torch.tensor(colours)).reshape(mIF.shape[1],mIF.shape[2],3)
    else:
        colour_render = mIF
    colour_render = torch.clamp_(colour_render, 0, 1)
    colour_render = _move_channel_axis(colour_render,to_back = True).detach().numpy()*255
    return colour_render.astype(np.uint8)





from instanseg.utils.utils import show_images, _move_channel_axis
import torchvision
import torchvision.transforms.functional as TF
import torch
import random
import numpy as np
from torchvision.transforms import RandomCrop, Resize, RandomPerspective
from monai.transforms import RandGaussianNoise, AdjustContrast
from instanseg.utils.utils import percentile_normalize, generate_colors
import warnings

import time
import fastremap

time_dict = {}


def measure_time(f):
    def timed(*args, **kw):
        ts = time.time()
        result = f(*args, **kw)
        te = time.time()
        if f.__name__ in time_dict.keys():
            time_dict[f.__name__] += te - ts
        else:
            time_dict[f.__name__] = te - ts
        # print((f.__name__, ))
        return result

    return timed


def measure_average_instance_area(lab):
    if torch.is_tensor(lab):
        if (len(torch.unique(lab) - 1)) > 0:
            return torch.sum((lab > 0).float()) / (len(torch.unique(lab) - 1))
        else:
            return 0
    else:
        if (len(np.unique(lab) - 1)) > 0:
            return np.sum((lab > 0)) / (len(np.unique(lab) - 1))
        else:
            return 0


def resize_lab(lab, size, double_check=False):
    """This function resizes a label to required instance area. It does so by first measuring the average instance area and assumes instances are roughly circular"""
    avg_size = measure_average_instance_area(lab)

    if avg_size == 0:
        return list(np.array(lab[0].shape))

    radius_orig = np.sqrt(avg_size / 3.14)  # Assuming labels are roughly circular
    radius_requested = np.sqrt(size / 3.14)

    # print(avg_size,size,radius_orig,radius_requested)

    ratio = radius_orig / radius_requested

    shape = np.array(lab[0].shape)

    if double_check:
        resized_lab = Resize(size=list(((shape / ratio).int())), antialias=True,
                             interpolation=torchvision.transforms.InterpolationMode.NEAREST)(lab[None,]).to(lab.dtype)
        assert np.isclose(measure_average_instance_area(resized_lab), size, rtol=0.6)

    return list(((shape / ratio).int()))


def generate_random_label_area(min=30, max=30): 
    from scipy.stats import skewnorm
    a = 5
    # mean, var, skew, kurt = skewnorm.stats(a, moments='mvsk')
    mean = 0.7824
    r = skewnorm.rvs(a, size=1)
    r = (r - mean) + 1
    r = np.clip(((r - mean) + 0.2) * 500, min, max)
    return r



def get_marker_location(meta):
    from instanseg.utils.augmentation_config import markers_info
    stains = [channel_str.split(" ")[0] for channel_str in meta['channel_names']]
    subcellular_location = ["N/A" if channel.upper() not in markers_info.keys() else markers_info[channel.upper()]['Subcellular Location'] for channel in stains]
    meta["subcellular_location"] = subcellular_location
    return meta
    
class Augmentations(object):
    def __init__(self, augmentation_dict={}, shape=(256, 256), dim_in=3,
                 nuclei_channel=None, debug=False, modality=None, cells_and_nuclei=False, target_segmentation="N",channel_invariant = False):
        self.debug = debug
        self.shape = shape
        self.augmentation_dict = augmentation_dict
        self.modality = modality
        self.cells_and_nuclei = cells_and_nuclei
        self.target_segmentation = target_segmentation
        self.dim_in = dim_in  # Note, this is the number of input channels to the model, not the number of channels in the raw image. (Can be 'None' for channel invariant models)
        self.nuclei_channel = nuclei_channel  # The channel that contains the nuclei. (Can be 'None' for brightfield images or if in image metadata)
        self.channel_invariant = channel_invariant
    def to_tensor(self, image, labels=None, normalize=False, amount=None, metadata=None):


        if isinstance(image, np.ndarray):
            if self.debug:
                orig = image.copy()

            if np.issubdtype(image.dtype, np.integer):
                image = image.astype(np.float32)
            if normalize:
                image, _ = self.normalize(image)

            out = torch.tensor(_move_channel_axis(image), dtype=torch.float32)

        elif isinstance(image, torch.Tensor):
            if self.debug:
                orig = torch.clone(image)
            out = _move_channel_axis(image).float()
        
            if normalize:
                out, _ = self.normalize(out)

        if labels is not None:
            if isinstance(labels, np.ndarray):
                labels = np.atleast_3d(labels)
                labels = _move_channel_axis(labels)
                for n, lab in enumerate(labels):

                    if not (lab == -1).any():  # convention is to skip labels with a value of -1
                        labels[n] = fastremap.renumber(lab)[0]
                    else:
                        labels[n] = lab

                labels = torch.tensor(labels.astype(np.int32), dtype=torch.int32)
            labels = torch.atleast_3d(labels)
            labels = _move_channel_axis(labels)

        out = out.squeeze()
        out = _move_channel_axis(torch.atleast_3d(out))

    
        if self.debug:
            print("Tensor")
            show_images([orig, out], titles=["Original", "Transformed"])
        return out, labels

    def normalize(self, image: torch.Tensor, labels=None, amount: float = 0., subsampling_factor: int = 1,
                  percentile=0.1, metadata=None):
        out = percentile_normalize(image, subsampling_factor=subsampling_factor, percentile=percentile)

        return out, labels
    

    def extract_hematoxylin_stain(self, image: torch.Tensor, labels=None, amount=0, metadata=None):
        # image should be 3 channel RGB between 0 and 255 (float32)
        if metadata is not None and metadata["image_modality"] != "Brightfield":
            return image, labels

        import torchstain
        if self.debug:
            orig = torch.clone(image)

        tensor = (image / (image.max() + 0.001)) * 255

        tensor = torch.clamp(tensor, 0., 255.)

        normalizer = torchstain.normalizers.MacenkoNormalizer(backend='torch')

        try:

            normalizer.HERef += (torch.rand_like(normalizer.HERef) - 0.5) * normalizer.HERef * amount
            normalizer.maxCRef += (torch.rand_like(normalizer.maxCRef) - 0.5) * normalizer.maxCRef * amount
            norm, H, E = normalizer.normalize(I=tensor, stains=True, Io=240)
        except:
            return image,labels

        out = _move_channel_axis(H) / 255.

        if self.debug:
            print("Stain separation")
            show_images([orig, out], titles=["Original", "Transformed"])

        return out, labels

    def normalize_HE_stains(self, image: torch.Tensor, labels=None, amount=0, metadata=None):
        # image should be 3 channel RGB between 0 and 255 (float32)
        if metadata is not None and metadata["image_modality"] != "Brightfield":
            return image, labels
        
        assert image.shape[0] == 3
        
        import torchstain
        if self.debug:
            orig = torch.clone(image)

        tensor = (image / (image.max() + 0.001)) * 255

        tensor = torch.clamp(tensor, 0., 255.)

        normalizer = torchstain.normalizers.MacenkoNormalizer(backend='torch')

        try:
            normalizer.HERef += (torch.rand_like(normalizer.HERef) - 0.5) * normalizer.HERef * amount
            normalizer.maxCRef += (torch.rand_like(normalizer.maxCRef) - 0.5) * normalizer.maxCRef * amount
            norm, _, _ = normalizer.normalize(I=tensor, stains=False, Io=240, beta=0.15)
        except:
            return image,labels


        out = _move_channel_axis(norm) / 255.

        if self.debug:
            print("Stain normalization")
            show_images([orig, out], titles=["Original", "Transformed"])

        return out, labels

    def randomJPEGcompression(self, image, labels=None, amount=0, metadata=None):
        """ This function applies random JPEG compression to an image. It is used to simulate the effect of JPEG compression on the image.

        :param image: A tensor of shape C,H,W
        :param labels: any labels associated with the image
        """

        if image.shape[0] != 3:  # Not implemented for >3 channels
            return image, labels

        import io
        from PIL import Image
        from torchvision import transforms
        import torchvision.transforms.functional as F
        import random

        if self.debug:
            orig = torch.clone(image)  # .copy()
        # Convert tensor to PIL image

        image = image.clamp(0, 1) * 255
        image = image.byte()


        image = F.to_pil_image(image)

        # # Apply random JPEG compression
        qf = random.randrange(int(100 * (1 - amount) - 1), 100)
        outputIoStream = io.BytesIO()
        image.save(outputIoStream, "JPEG", quality=qf, optimize=True)
        outputIoStream.seek(0)

        # # Convert compressed image back to Torch tensor
        out = transforms.ToTensor()(Image.open(outputIoStream))

        if self.debug:
            print("JPEG")
            show_images([orig, out, orig - out], titles=["Original", "Transformed"])

        return out, labels

    def brightness_augment(self, image, labels=None, amount=0, metadata=None):
        if self.debug:
            orig = torch.clone(image)  # .copy()
        rand = 1 + ((torch.rand(len(image)) - 0.5) * amount)
        out = (image.permute(1, 2, 0) * rand).permute(2, 0, 1)
        if self.debug:
            print("Brightness")
            show_images([orig, out], titles=["Original", "Transformed"])
        return out, labels

    def RandGaussianNoise(self, image, labels=None, amount=0, metadata=None):

        if self.debug:
            orig = torch.clone(image)  # .copy()

        out = RandGaussianNoise(prob=1, mean=0.0, std=amount / 5)(image)

        if self.debug:
            print("RandGaussianNoise")
            show_images([orig, out], titles=["Original", "Transformed"])
        return out, labels

    def HistogramNormalize(self, image, labels=None, amount=0, metadata=None):
        from monai.transforms import RandStdShiftIntensity

        if self.debug:
            orig = torch.clone(image)  # .copy()

        normalizer = RandStdShiftIntensity(10)

        out = torch.stack([normalizer(c) for c in image])

        if self.debug:
            print("RandStdShiftIntensity")
            show_images([orig, out], titles=["Original", "Transformed"])
        return out, labels
    

    def AdjustContrast(self, image, labels=None, amount=0, metadata=None):

        if self.debug:
            orig = torch.clone(image)  # .copy()

        out = AdjustContrast(gamma=amount)(image)

        if self.debug:
            print("AdjustContrast")
            show_images([orig, out], titles=["Original", "Transformed"])
        return out, labels
    
    def flips(self, image, labels, amount=0, metadata=None):

        amount = 0.5

        if self.debug:
            orig = torch.clone(image)  # .copy()
        if random.random() > (1 - amount):
            image = TF.hflip(image)
            labels = TF.hflip(labels)
        if random.random() > (1 - amount):
            image = TF.vflip(image)
            labels = TF.vflip(labels)
        out = image
        if self.debug:
            print("Flips")
            show_images([orig, out])
        return out, labels

    def rotate(self, image, labels, amount=0, metadata=None):

        if self.debug:
            orig = torch.clone(image)

        angle = int(np.random.choice([180, 90, 270, 0]))
        out = TF.rotate(image, angle)
        labels = TF.rotate(labels, angle)

        if self.debug:
            print("Rotate")
            show_images([orig, out])
        return out, labels

    def adjust_hue(self, image, labels, amount=0, metadata=None):

        assert image.shape[0] == 3

        if metadata is not None and metadata["image_modality"] != "Brightfield":
            return image, labels

        if self.debug:
            orig = torch.clone(image)

        out = TF.adjust_hue(image, hue_factor=(torch.rand(1) - 0.5) * amount * 0.5)

        out = torch.nan_to_num(out, nan=0)  # This occasionally produces nan values, which we replace with 0

        if self.debug:
            print("Adjust Hue")
            show_images([orig, out], titles=["Original", "Transformed"])
        return out, labels

    def perspective(self, image, labels, amount=0, metadata=None):

        if self.debug:
            orig = torch.clone(image)  # .copy()
        perspective_transformer = RandomPerspective(distortion_scale=amount / 2, p=1.0,
                                                    interpolation=torchvision.transforms.InterpolationMode.NEAREST,
                                                    fill=float(image.max() if metadata[
                                                                                  "image_modality"] == "Brightfield" else image.min()))
        state = torch.get_rng_state()

        out = perspective_transformer(image)

        perspective_transformer = RandomPerspective(distortion_scale=amount / 2, p=1.0,
                                                    interpolation=torchvision.transforms.InterpolationMode.NEAREST,
                                                    fill=0)

        torch.set_rng_state(state)
        labels = perspective_transformer(labels)

        if self.debug:
            print("perspective")
            show_images([orig, out, labels], )
        return out, labels

    def invert(self, image, labels, amount=0, metadata=None):

        out = 1 - image

        if self.debug:
            print("Inversion")
            show_images([image, out])
        return out, labels


    # @measure_time
    def pseudo_brightfield(self, image, labels=None, amount=0, c_nuclei=0, metadata=None, random_seed=None):

        if self.debug:
            orig = torch.clone(image)  # .copy()
        orig_dtype = image.dtype

        if random_seed is not None:
            np.random.seed(random_seed)

        if metadata is not None and metadata["image_modality"] != "Fluorescence":
            return image, labels
        
        image , _ , c_nuclei= self.extract_nucleus_and_cytoplasm_channels(image, labels, c_nuclei, metadata)

        image = percentile_normalize(image, 0.01, subsampling_factor=5)

        stains = {
            'hematoxylin': torch.Tensor([0.65, 0.70, 0.29]),
            'dab': torch.Tensor([0.27, 0.57, 0.78]),
            'eosin': torch.Tensor([0.2159, 0.8012, 0.5581])
        }

        assert c_nuclei is not None
        im_nuclei = image[c_nuclei]

        my_stains = [torch.clone(stains['hematoxylin'])]
        channels = [im_nuclei]

        if np.random.random() > 0.1:
            my_stains.append(torch.clone(stains['dab']))

            all_channels = np.arange(len(image))
            c_other = np.random.choice(all_channels[all_channels != c_nuclei])
            channels.append(image[c_other])

        for ii in range(len(my_stains)):
            stain = my_stains[ii]
            stain += np.random.random(3) / 4
            stain /= np.linalg.norm(stain)
            stain_scale = 1 + (np.random.random() - 0.5) / 5
            my_stains[ii] = stain * stain_scale

        im_output = 0
        channels = [temp for temp in channels]

        for stain, im_channel in zip(my_stains, channels):
            stain = stain.reshape((3, 1, 1))

            im_output += im_channel * stain * float(np.random.randint(3, 5))

        im_output = torch.exp(-np.log(10.0) * im_output)
        im_output = torch.clip(im_output, 0, 1)

        if self.debug:
            print("Eosin")
            show_images([orig, im_output], titles=["Original", "Transformed"])

        return torch.Tensor(im_output).to(orig_dtype), labels

    def colourize(self, image, labels=None, amount=0, c_nuclei=-1, random_seed=None, metadata=None):

        # Expects a 3D tensor C,H,W

        if self.debug:
            orig = torch.clone(image)  # .copy()

        if random_seed is not None:
            np.random.seed(random_seed)

        if metadata is not None and metadata["image_modality"] != "Fluorescence":
            return image, labels
        # colours = [[1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 1, 1], [1, 0, 1]]

        if image.shape[0] == 3:
            return image, labels
        
        if c_nuclei is None:
            c_nuclei = 0

        colours = generate_colors(num_colors=image.shape[0])
        np.random.shuffle(colours)
        if np.min(image.shape) == 1:
            c_nuclei = 0  # If we have a greyscale image, we want to colourize the one and only nuclei channel
        
        coloured_image = image[c_nuclei][None,] * torch.Tensor([0, 0, 1])[:, None, None]
        for i, image_channel in enumerate(image[np.arange(len(image)) != c_nuclei]):
            colour = colours[i]
            # greyscale_image_channel = image_channel[None,].expand(3, -1, -1)
            coloured_image_channel = image_channel * torch.Tensor(colour)[:, None, None]
            coloured_image += coloured_image_channel
        assert not coloured_image.isnan().any()
        image = percentile_normalize(coloured_image, 0.1, subsampling_factor=5)
        if self.debug:
            print("Colourize")
            show_images([orig, image], titles=["Original", "Transformed"])
        assert not coloured_image.isnan().any()
        return image, labels

    def pseudo_imc(self, image, labels, amount=0, metadata=None):

        # Expects a 3D tensor C,H,W and labels C,H,W
        

        if metadata is not None and metadata["image_modality"] != "Fluorescence":
            return image, labels

        if self.debug:
            orig = torch.clone(image)  # .copy()

        original_shape = image.shape

        if np.min(torch.squeeze(
                image).shape) == 3:  # Don't want to run the risk of having a brightfield image in the IMC pipeline.
            if self.debug:
                warnings.warn("Possible Brightfield image in IMC pipeline, ignoring.")
                show_images([orig, image], titles=["Original", "Transformed"])

            return image, labels

        shape = resize_lab(labels,
                           generate_random_label_area(min=30, max=50))  # Rough size for the nuclei in IMC images.

        resized_data = Resize(size=shape, antialias=True)(image)

        resized_data[resized_data < 0] = 0

        amount = 30 - (
                amount * 27)  # Assuming we have amount between [0,1], we want to map that to something like [30,3]

        resized_data = torch.poisson(resized_data * amount)

        resized_data = percentile_normalize(resized_data, percentile=0.1, subsampling_factor=3)

        image = Resize(size=original_shape[1:], interpolation=np.random.choice(
            [torchvision.transforms.InterpolationMode.NEAREST, torchvision.transforms.InterpolationMode.BILINEAR]),
                       antialias=True)(
            resized_data)

        if self.debug:
            print("Pseudo_imc")
            show_images([orig, image], titles=["Original", "Transformed"])

        return image, labels
    
    def channel_shuffle(self, image, labels=None, amount=0, metadata=None):
        if self.debug:
            orig = torch.clone(image)
        channels = torch.randperm(image.shape[0])
        out = image[channels]
        if self.debug:
            print("Channel shuffle")
            show_images([orig, out], titles=["Original", "Transformed"])
        return out, labels
    
    def add_noisy_channels(self, image, labels=None, max_channels = 30, amount=0, metadata=None):
        if self.debug:
            orig = torch.clone(image)

        new_channels_num = np.random.randint(1,max_channels)

        new_channels =np.random.choice(range(image.shape[0]),new_channels_num,replace = True)

        out = image[new_channels]
        out[out<0]=0
        amount = np.random.randint(2, 30 * amount)

        channel_weights = torch.randint(1,amount,size = [new_channels_num,1,1],dtype = torch.float32)


        out = torch.poisson(out * channel_weights) / (channel_weights*2) 

        out = torch.cat((image,out),dim=0)

        channels = torch.randperm(out.shape[0])
        out = out[channels]


        return out, labels

    # @measure_time
    def add_gradient(self, image, labels=None, amount=0, metadata=None):
        if self.debug:
            orig = torch.clone(image)  # .copy()
        _, h, w = image.shape
        xs = torch.linspace(0, 0.5, steps=w)
        ys = torch.linspace(0, 0.5, steps=h)
        x, y = torch.meshgrid(xs, ys, indexing='xy')
        if random.random() > 0.5:
            x = TF.hflip(x)
        if random.random() > 0.5:
            y = TF.hflip(y)
        if random.random() > 0.5:
            x = TF.vflip(x)
        if random.random() > 0.5:
            y = TF.vflip(y)
        out = (amount * (np.random.random() * x + np.random.random() * y)) * image.max() + image
        if self.debug:
            print("Gradient")
            show_images([orig, out], titles=["Original", "Transformed"])

        return torch.Tensor(out), labels

    # @measure_time

    def channel_subsample(self, image, labels=None, max_channels=None, c_nuclei=None, min_channels=1, metadata=None):
        channel_num = image.shape[0]

        if channel_num > max_channels:
            if c_nuclei is None or c_nuclei > channel_num:
                slice = np.random.choice(np.arange(channel_num), max_channels, replace=False)
            else:
                if max_channels == 1:
                    slice = [c_nuclei]
                    c_nuclei = 0
                else:

                    slice = np.random.choice(np.arange(channel_num)[np.arange(channel_num) != c_nuclei],
                                             np.random.randint(min_channels - 1, max_channels - 1), replace=False)

                    np.random.shuffle(slice)
                    slice = np.append(slice, c_nuclei)

                    c_nuclei = np.where(slice == c_nuclei)[0][0]

            image = image[slice]

        return image, labels, c_nuclei
    
    def channel_suppress(self, image, labels=None, amount = None, metadata=None):

        slice = torch.rand(image.shape[0]) < (1- amount)

        if torch.sum(slice) == 0:
            slice[np.random.randint(0,image.shape[0])] = True

        image = image[slice]

        return image, labels

    def extract_nucleus_and_cytoplasm_channels(self, image, labels=None, c_nuclei=None, metadata=None, amount=None):
        channel_num = image.shape[0]

        if metadata is None or "subcellular_location" not in metadata.keys() or len(
                metadata["subcellular_location"]) != channel_num or c_nuclei is None:
            return image, labels, c_nuclei

        cytoplasm_channel_ids = ["Cytoplasm" in x or "Membrane" in x or "N/A" in x for x in
                                 metadata["subcellular_location"]]
        nuclei_channels_ids = c_nuclei

        if len(cytoplasm_channel_ids) == 0:
            return image, labels, c_nuclei

        cytoplasm_channels = image[cytoplasm_channel_ids].sum(0)  # Sum projection of the cytoplasm channels
        nuclei_channels = image[nuclei_channels_ids]

        image = torch.stack([nuclei_channels, cytoplasm_channels])

        c_nuclei = 0

        return image, labels, c_nuclei

    
    def torch_rescale(self, image, labels=None, amount=0, current_pixel_size=None, requested_pixel_size=None, crop=True,
                      random_seed=None, metadata=None, modality = None):

    

        if random_seed is not None:
            torch.manual_seed(random_seed)

        if labels is not None:
            assert image.shape[-2:] == labels.shape[-2:]


        shape = (torch.tensor(image.shape[-2:]) * (1 + ((torch.rand(1) - 0.5) * amount))).int().tolist()

        if metadata is not None:
            if current_pixel_size is None and "pixel_size" in metadata.keys():
                current_pixel_size = metadata["pixel_size"]
            if modality is None and "image_modality" in metadata.keys():
                modality = metadata["image_modality"]
            else:
                modality = None

        if modality is None:
            modality = "Brightfield"
        #    print("Modality not specified in metadata or in function call, assuming Brightfield")


        if current_pixel_size is not None and requested_pixel_size is not None:
            scale = (current_pixel_size / requested_pixel_size)
            shape = (torch.tensor(image.shape[-2:]) * scale).int().tolist()

        resized_data = Resize(size=shape, antialias=True)(image)



        if labels is not None:

            resized_labels = Resize(size=shape, interpolation=torchvision.transforms.InterpolationMode.NEAREST)(labels)

        while self.shape is not None and np.any(np.array(resized_data[0].shape) < self.shape[0]) and crop:
            pad = int((self.shape[0] - min(resized_data[0].shape)) / 2) + 3
            pad = torch.Tensor([pad, resized_data.shape[1], resized_data.shape[2]]).min().int() - 1

            

            resized_data = torch.nn.functional.pad(resized_data, (pad, pad, pad, pad), mode='constant',
                                                   value=image.max() if modality == "Brightfield" else resized_data.min())

            if labels is not None:
                resized_labels = torch.nn.functional.pad(resized_labels, (pad, pad, pad, pad), mode='constant', value = min(resized_labels.min(),0)).to(
                    labels.dtype)
                
              
        if not crop:
            if labels is None:
                return resized_data, None
            else:
                return resized_data, resized_labels

        cropper = RandomCrop(size=self.shape)
        (i, j, h, w) = cropper.get_params(resized_data, output_size=self.shape)
        out_image = resized_data[:, i:i + h, j:j + w]

        if labels is not None:
            out_labels = resized_labels[:, i:i + h, j:j + w]

            assert out_image.shape[-2:] == self.shape and out_labels.shape[-2:] == self.shape

        if self.debug:
            print("Rescaling")
            show_images([image[0], labels, out_image[0], out_labels],
                        titles=["Source", "Target", "Resize source", "Resized target"], n_cols=2, axes=True)
            
        

        if labels is None:
            return out_image.float(), None
        else:
            return out_image.float(), out_labels #torch.Tensor(out_labels)  # .short()
        
        
    def duplicate_grayscale_channels(self, image, labels, metadata=None):
        if self.debug:
            orig = torch.clone(image)
        if image.shape[0] == 1 and self.dim_in != 1 and not self.channel_invariant:
            image = image.repeat(self.dim_in, 1, 1)
        elif image.shape[0] != self.dim_in and not self.channel_invariant:
            raise ValueError(
                "Image has {} channels, but model expects {}, check the augmentations pipeline!".format(image.shape[0],
                                                                                                        self.dim_in))

        if self.debug:
            print("Duplicate channels")
            show_images([orig, image], titles=["Original", "Transformed"])
        return image, labels

    def __call__(self, image, labels, meta=None):

        from instanseg.utils.utils import _estimate_image_modality

        if self.modality is None:
            if meta is not None and "image_modality" in meta.keys():
                if meta["image_modality"] in ["Brightfield", "Chromogenic"]:
                    observed_modality = "Brightfield"
                else:
                    observed_modality = meta["image_modality"]
            else:
                observed_modality = _estimate_image_modality(image, labels)
            modality = observed_modality
        else:
            modality = self.modality

        if self.debug:
            print("Observed modality:", modality)


        augmentation_dict = self.augmentation_dict[modality]

        if self.debug:
            print(augmentation_dict.keys())

        if meta is not None:
            if "nuclei_channels" in meta.keys():
                channels_nuclei = meta["nuclei_channels"]
                c_nuclei = np.random.choice(channels_nuclei)

            else:
                c_nuclei = self.nuclei_channel

            if "pixel_size" in meta.keys():
                pixel_size = meta["pixel_size"]
            else:
                pixel_size = None

        if meta is not None and "channel_names" in meta.keys():
            from instanseg.utils.augmentation_config import markers_info
            stains = [channel_str.split(" ")[0] for channel_str in meta['channel_names']]
            subcellular_location = [
                "N/A" if channel.upper() not in markers_info.keys() else markers_info[channel.upper()][
                    'Subcellular Location'] for channel in stains]
        else:
            subcellular_location = ["N/A" for _ in range(image.shape[0])]

        metadata = {"image_modality": modality, "nuclei_channels": c_nuclei, "pixel_size": pixel_size,
                    "subcellular_location": subcellular_location}

        has_been_normalized = False

        for augmentation, values in augmentation_dict.items():

            if np.random.random() < values[0]:

                if augmentation in ["normalize_HE_stains", "extract_hematoxylin_stain", "normalize", "pseudo_background"]:
                    if not has_been_normalized:
                        if augmentation != "normalize":
                            _, amount = values
                        image, labels = getattr(self, augmentation)(image, labels, amount=amount, metadata=metadata)
                        has_been_normalized = True

                    else:
                        pass

                elif augmentation == "pseudo_brightfield":
                    image, labels = self.pseudo_brightfield(image, labels, c_nuclei=c_nuclei, metadata=metadata)
                    metadata["image_modality"] = "Brightfield"

                elif augmentation == "channel_subsample":
                    _, (min_channels, max_channels) = values
                    image, labels, c_nuclei = self.channel_subsample(image, labels, max_channels=max_channels + 1,
                                                                     c_nuclei=c_nuclei, min_channels=min_channels,
                                                                     metadata=metadata)
                elif augmentation == "extract_nucleus_and_cytoplasm_channels":
                    image, labels, c_nuclei = self.extract_nucleus_and_cytoplasm_channels(image, labels,
                                                                                          c_nuclei=c_nuclei,
                                                                                          metadata=metadata)

                elif augmentation == "torch_rescale":
                    _, requested_pixel_size, amount = values
                    image, labels = self.torch_rescale(image, labels, current_pixel_size=None,
                                                       requested_pixel_size=requested_pixel_size, amount=amount,
                                                       metadata=metadata)

                elif augmentation == "colourize":
                    _, amount = values
                    image, labels = self.colourize(image, labels, c_nuclei=c_nuclei, metadata=metadata)


                elif augmentation == "add_noisy_channels":
                    _, max_channels = values
                    image, labels = self.add_noisy_channels(image, labels, metadata=metadata, max_channels=max_channels, amount = 0.5)

                else:
                    p, *rest = values
                    amount = rest[0] if len(rest) == 1 else None

                    image, labels = getattr(self, augmentation)(image, labels, amount=amount, metadata=metadata)

                assert not image.isnan().any()
        

        image, labels = self.duplicate_grayscale_channels(image,
                                                          labels,
                                                          metadata=metadata)  # This will only catch single channel images fed to a multi channel network and duplicate the channel if required.

        if image.var() > 1e2:
            image = torch.clip(image, min=-1, max=5)
            warnings.warn("Warning, variance of image is very high, check augmentations")

        return image, labels


if __name__ == "__main__":
    from instanseg.utils.augmentation_config import get_augmentation_dict

    augmentation_dict = get_augmentation_dict(nuclei_channel=6, dim_in=3, amount=0.5,augmentation_type="heavy")['train']

    import tifffile

    img = tifffile.imread(r"../examples/LuCa1.tif")[:,:512,:512]
    label = tifffile.imread(r"../examples/LuCa1_label.tif")[:512,:512]
    meta = {"image_modality": "Fluorescence", "nuclei_channels": [7]}


    Augmenter = Augmentations(augmentation_dict=augmentation_dict, debug=False,dim_in = None)

    show_images([Augmenter(img, label, meta)[0] for i in range(30)])
